#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>

#include <sys/mman.h>  // for mmap()
#include <fcntl.h>     // for open() et al.

#include "bstd.h"

void die_perror(const char *err, ...) {
  va_list params;

  va_start(params, err);
  vfprintf(stderr, err, params);
  fputc('\n', stderr);
  va_end(params);
  perror("errmsg");
  exit(1);
}

void bmkdir(const char *pathname) {
  if (mkdir(pathname, 0777)) die_perror("mkdir %s", pathname);
}

void bchdir(const char *pathname) {
  if (chdir(pathname)) die_perror("chdir %s", pathname);
}

void bfclose(FILE *fp) {
  if (fclose(fp)) die_perror("fclose");
}

FILE *bfopen(const char *path, const char *mode) {
  FILE *fp = fopen(path, mode);
  if (!fp) die_perror("fopen %s", path);
  return fp;
}

time_t btouch_get_mtime(const char *file) {
  // If file exists, check it is regular?
  int fd = open(file, O_CREAT | O_WRONLY, 0666);
  if (fd < 0) die_perror("open");
  struct stat st[1];
  if (fstat(fd, st)) die_perror("fstat");
  close(fd);
  return st->st_mtime;
}

void bstrtofile(const char *pathname, const char *str) {
  FILE *fp = bfopen(pathname, "w");
  if (EOF == fputs(str, fp)) die_perror("fputs");
  bfclose(fp);
}

char *bfiletostr(int *len, const char *pathname) {
  int fd = bopen(pathname, O_RDONLY);
  int n = bfd_size_of(fd);
  char *res = malloc(n + 1);
  if (n != read(fd, res, n)) die("read %s", pathname);
  res[n] = '\0';
  if (len) *len = n;
  return res;
}

void bstrappendfile(const char *pathname, const char *str) {
  FILE *fp = bfopen(pathname, "a");
  if (EOF == fputs(str, fp)) die_perror("fputs");
  bfclose(fp);
}

void bunlink(const char *pathname) {
  if (unlink(pathname)) die_perror("unlink");
}

void bfwrite_bytes(const void *ptr, size_t n, FILE *stream) {
  size_t r = fwrite(ptr, 1, n, stream);
  if (r < n) die_perror("fwrite");
}

static int bmode_from_stat(mode_t st_mode) {
  if (S_ISREG(st_mode)) {
    return bmode_reg;
  }
  if (S_ISLNK(st_mode)) {
    return bmode_link;
  }
  if (S_ISDIR(st_mode)) {
    return bmode_dir;
  }
  return bmode_unknown;
}

int blstat(const char *pathname) {
  struct stat st[1];
  if (lstat(pathname, st)) die_perror("lstat %s", pathname);
  return bmode_from_stat(st->st_mode);
}

int direxists(const char *pathname) {
  struct stat buf;
  if (lstat(pathname, &buf)) return 0;
  return bmode_dir == bmode_from_stat(buf.st_mode);
}

int fileexists(const char *pathname) {
  struct stat buf;
  if (lstat(pathname, &buf)) return 0;
  return bmode_reg == bmode_from_stat(buf.st_mode);
}

void bopendir(const char *pathname, void (*callback)(const char *)) {
  DIR *d = opendir(pathname);
  if (!d) {
    fprintf(stderr, "not opening %s: %s\n", pathname, strerror(errno));
    return;
  }
  struct dirent *de;
  while ((de = readdir(d))) {
    char *filename = de->d_name;
    if (('.' == filename[0]) &&
        ('\0' == filename[1] ||
	 !strcmp(filename + 1, ".") ||
	 !strcmp(filename + 1, "gg"))) {
      continue;
    }
    callback(filename);
  }
  closedir(d);
}

void die(const char *err, ...) {
  va_list params;

  va_start(params, err);
  vfprintf(stderr, err, params);
  fputc('\n', stderr);
  exit(1);
  va_end(params);
}

char *bstrdup(const char *s) {
  char *res = strdup(s);
  if (!res) die_perror("strdup");
  return res;
}

size_t bfd_size_of(int fd) {
  struct stat st[1];
  if (fstat(fd, st)) die_perror("fstat");
  return (size_t) st->st_size;
}

int bopen(const char *pathname, int flags) {
  int fd = open(pathname, flags, 0644);
  if (fd < 0) die_perror("open %s", pathname);
  return fd;
}

int bopen_write(const char *pathname) {
  return bopen(pathname, O_WRONLY | O_CREAT);
}

void *map_file_to_memory(size_t *len, const char *path) {
  int fd = bopen(path, O_RDONLY);
  *len = bfd_size_of(fd);
  if (!*len) {
    close(fd);
    return NULL;
  }
  void *map = mmap(NULL, *len, PROT_READ, MAP_PRIVATE, fd, 0);
  if (MAP_FAILED == map) die("mmap failed");
  close(fd);
  return map;
}

// malloc-like call, but creates and returns memory map.
void *mmap_alloc(size_t len) {
  void *map = mmap(NULL, len, PROT_READ | PROT_WRITE,
      MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
  if (MAP_FAILED == map) die("mmap failed");
  return map;
}

void unmap_memory(void *map, size_t len) {
  if (map) munmap(map, len);
}

void bfputc(char c, FILE *fp) {
  if (EOF == fputc(c, fp)) die_perror("fputc");
}

void bmake_executable(const char *file) {
  struct stat st[1];
  if (lstat(file, st)) die_perror("bmake_exectuable %s: lstat", file);
  if (chmod(file, 0111 | st->st_mode)) {
    die_perror("bmake_exectuable %s: chmod", file);
  }
}

int prefixcmp(const char *str, const char *prefix) {
  for (; ; str++, prefix++) {
    if (!*prefix) return 0;
    else if (*str != *prefix) {
      return (unsigned char)*prefix - (unsigned char)*str;
    }
  }
}

