#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <unistd.h>
#include "bstd.h"
#include "id.h"
#include "db_t.h"

// Filesystem object database: all information stored in .gg directory.

struct ggdb_s {
  char *path;
  int pathlen;
};
typedef struct ggdb_s *ggdb_ptr;

/*
static char *cleanup_path(char *messy_path) {
  if (!memcmp(messy_path, "./", 2)) {
    char *path = messy_path + 2;
    while (*path == '/') path++;
    char *res = malloc(strlen(path) + 1);
    strcpy(res, path);
    free(messy_path);
    return res;
  }
  return messy_path;
}
*/

#include <stdarg.h>
static char *gg_pathf(ggdb_ptr gg, const char *fmt, ...) {
  const char *gg_dir = gg->path;
  char *pathname = malloc(PATH_MAX);
  va_list args;
  unsigned len;

  len = strlen(gg_dir);
  // if (len > PATH_MAX-100) return bad_path;
  memcpy(pathname, gg_dir, len);
  if (len && gg_dir[len - 1] != '/')
	  pathname[len++] = '/';
  va_start(args, fmt);
  len += vsnprintf(pathname + len, PATH_MAX - len, fmt, args);
  va_end(args);
  // if (len >= PATH_MAX) return bad_path;
  // return cleanup_path(pathname);
  return pathname;
}

static void gg_delete(db_ptr db) {
  ggdb_ptr gg = db_data(db);
  free(gg->path);
  free(db);
}

static int gg_reset(db_ptr db) {
  ggdb_ptr gg = db_data(db);
  char s[gg->pathlen + 16];
  strcpy(s, gg->path);
  bmkdir(gg->path);
  strcpy(s + gg->pathlen, "objects");
  bmkdir(s);
  bmkdir(".gg/refs");
  bmkdir(".gg/refs/heads");
  return 0;
}

static void *gg_mmap_object_raw(size_t *len, db_ptr db, const ID id) {
  ggdb_ptr gg = db_data(db);
  char obj[gg->pathlen + PATH_ID_LEN];
  cat_path_id(obj, gg->path, id);
  void *res = map_file_to_memory(len, obj);
  if (!res) die("invalid object");
  return res;
}

static int gg_has_branch(db_ptr db, const char *s) {
  ggdb_ptr gg = db_data(db);
  char *branch = gg_pathf(gg, "refs/heads/%s", s);
  int res = fileexists(branch);
  free(branch);
  return res;
}

static void *gg_mmap_ref_raw(size_t *len, db_ptr db, const char *ref) {
  ggdb_ptr gg = db_data(db);
  char s[gg->pathlen + strlen(ref) + 1];
  strncpy(s, gg->path, gg->pathlen);
  strcpy(s + gg->pathlen, ref);
  char *cp = strrchr(s, '\n');
  if (cp) *cp = '\0';
  return map_file_to_memory(len, s);
}

static void gg_write_ref_raw(db_ptr db, const char *ref,
    void *buf, size_t len) {
  ggdb_ptr gg = db_data(db);
  char path[gg->pathlen + strlen(ref) + 16];
  strcpy(path, gg->path);
  strcat(path, ref);
  char *cp = strrchr(path, '\n');
  if (cp) *cp = '\0';

  char *slash;
  cp = path + gg->pathlen;
  while ((slash = strchr(cp, '/'))) {
    *slash = '\0';
    if (!direxists(path)) bmkdir(path);
    *slash = '/';
    cp = slash + 1;
  }
  int fd = bopen_write(path);
  write(fd, buf, len);  // TODO: Write bwrite().
  close(fd);
}

static int gg_has_object(db_ptr db, const ID id) {
  ggdb_ptr gg = db_data(db);
  char obj[gg->pathlen + PATH_ID_LEN];
  cat_path_id(obj, gg->path, id);
  return fileexists(obj);
}

static void gg_forall_branches(db_ptr db, void (*fn)(const char *ref)) {
  char s[PATH_MAX];
  int slen = 0, oldlen;
  s[0] = '\0';
  void walk(const char *path) {
    int bmode = blstat(path);
    switch(bmode) {
    case bmode_dir:
      oldlen = slen;
      if (slen) strcat(s, "/");
      strcat(s, path);
      slen = strlen(s);
      bchdir(path);
      bopendir(".", walk);
      bchdir("..");
      slen = oldlen;
      s[slen] = '\0';
      break;
    case bmode_reg:
      if (slen) strcat(s, "/");
      strcat(s, path);
      fn(s);
      s[slen] = '\0';
      break;
    default:
      fprintf(stderr, "ignoring bad branch %s", path);
      break;
    }
  }
  ggdb_ptr gg = db_data(db);
  char *ref_dir = gg_pathf(gg, "refs/heads/");
  bchdir(ref_dir);
  bopendir(".", walk);
  free(ref_dir);
}

static void gg_write_header_body_raw(db_ptr db, const ID id,
    const char *header, const unsigned char *buf, size_t len) {
  // TODO: Write to temp file and then move.
  ggdb_ptr gg = db_data(db);
  char obj[gg->pathlen + PATH_ID_LEN];
  cat_path_id(obj, gg->path, id);

  if (fileexists(obj)) return;

  char *dir = gg_pathf(gg, "objects/%02x", id[0]);
  if (!direxists(dir)) bmkdir(dir);
  free(dir);

  FILE *fp = bfopen(obj, "wb");

  if (fp) {
    fwrite(header, 1, strlen(header), fp);
    fwrite(buf, 1, len, fp);
    bfclose(fp);
  }
}

static void gg_appendlog(db_ptr db, const ID id, const char *msg) {
  ggdb_ptr gg = db_data(db);
  char *s = gg_pathf(gg, "log");
  FILE *fp = bfopen(s, "ab");
  id_fprint(fp, id);
  fprintf(fp, " %s\n", msg);
  bfclose(fp);
  free(s);
}

// db_new_gg() should be the only extern function.
db_ptr db_new_gg(const char *path) {
  db_ptr db = malloc(sizeof(*db));
  db->data = malloc(sizeof(struct ggdb_s));
  ggdb_ptr gg = db_data(db);
  gg->path = strdup(path);
  gg->pathlen = strlen(path);

  db->delme = gg_delete;
  db->reset = gg_reset;
  db->write_ref_raw = gg_write_ref_raw;
  db->mmap_ref_raw = gg_mmap_ref_raw;
  db->has_branch = gg_has_branch;
  db->forall_branches = gg_forall_branches;
  db->has_object = gg_has_object;
  db->mmap_object_raw = gg_mmap_object_raw;
  db->write_header_body_raw = gg_write_header_body_raw;
  db->appendlog = gg_appendlog;
  return db;
}

// TODO: Work out who should initialize the index.
void index_init(index_ptr index, const char *index_file);
void db_index_init(index_ptr index, db_ptr db) {
  ggdb_ptr gg = db_data(db);
  char *index_file = gg_pathf(gg, "index");
  index_init(index, index_file);
  free(index_file);
}
