#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <unistd.h>
#include "bstd.h"
#include "id.h"
#include "db_t.h"

// Filesystem object database with lazy fetching: information stored
// in .gg directory, fetched from URL when neeeded.

struct lazydb_s {
  char *path;
  int pathlen;
  char *tmpdir;
  int tmplen;
  char *url;
  int urllen;
};
typedef struct lazydb_s *lazydb_ptr;

static void lazy_delete(db_ptr db) {
  lazydb_ptr lazy = db_data(db);
  free(lazy->path);
  rmdir(lazy->tmpdir);
  free(lazy->tmpdir);
  free(db);
}

static int lazy_reset(db_ptr db) {
  lazydb_ptr lazy = db_data(db);
  char s[lazy->pathlen + 16];
  strcpy(s, lazy->path);
  bmkdir(lazy->path);
  strcpy(s + lazy->pathlen, "objects");
  bmkdir(s);
  bmkdir(".gg/refs");
  bmkdir(".gg/refs/heads");
  FILE *fp = fopen(".gg/lazy", "wb");
  fprintf(fp, "%s\n", lazy->url);
  bfclose(fp);
  return 0;
}

static void *mmap_download(size_t *len, lazydb_ptr lazy, const char *url) {
  char cmd[strlen(url) + 128];
  ID id;
  char file[lazy->tmplen + ID_SLEN + 2];
  id_string(id, url);
  sprintf(file, "%s/", lazy->tmpdir);
  id_sprint(file + lazy->tmplen + 1, id);
  sprintf(cmd, "curl %s -o %s", url, file);
  printf("command: %s\n", cmd);
  if (system(cmd)) die("system() error: %s", cmd);
  void *res = map_file_to_memory(len, file);
  unlink(file);
  return res;
}

static void lazy_write_header_body_raw(db_ptr db, const ID id,
    const char *header, const unsigned char *buf, size_t len) {
  // TODO: Write to temp file and then move.
  lazydb_ptr lazy = db_data(db);
  char obj[lazy->pathlen + PATH_ID_LEN];
  cat_path_id(obj, lazy->path, id);

  if (fileexists(obj)) return;

  char dir[lazy->pathlen + 16];
  sprintf(dir, "%sobjects/%02x", lazy->path, id[0]);
  if (!direxists(dir)) bmkdir(dir);

  FILE *fp = bfopen(obj, "wb");

  if (fp) {
    fwrite(header, 1, strlen(header), fp);
    fwrite(buf, 1, len, fp);
    bfclose(fp);
  }
}

static void *lazy_mmap_object_raw(size_t *len, db_ptr db, const ID id) {
  lazydb_ptr lazy = db_data(db);
  char obj[lazy->pathlen + PATH_ID_LEN];
  cat_path_id(obj, lazy->path, id);
  void *res;
  if (fileexists(obj)) {
    res = map_file_to_memory(len, obj);
    if (!res) die("invalid object");
  } else {
    char url[lazy->urllen + PATH_ID_LEN];
    cat_path_id(url, lazy->url, id);
    // TODO: Should download and rename, not download and copy.
    res = mmap_download(len, lazy, url);
    if (!res) die("invalid object or bad download");
    lazy_write_header_body_raw(db, id, "", res, *len);
  }
  return res;
}

static int lazy_has_branch(db_ptr db, const char *s) {
  lazydb_ptr lazy = db_data(db);
  char branch[lazy->pathlen + strlen(s) + 16];
  sprintf(branch, "%srefs/heads/%s", lazy->path, s);
  return fileexists(branch);
}

static void *lazy_mmap_ref_raw(size_t *len, db_ptr db, const char *ref) {
  lazydb_ptr lazy = db_data(db);
  char s[lazy->pathlen + strlen(ref) + 1];
  strncpy(s, lazy->path, lazy->pathlen);
  strcpy(s + lazy->pathlen, ref);
  char *cp = strrchr(s, '\n');
  if (cp) *cp = '\0';
  return map_file_to_memory(len, s);
}

static void lazy_write_ref_raw(db_ptr db, const char *ref,
    void *buf, size_t len) {
  lazydb_ptr lazy = db_data(db);
  char path[lazy->pathlen + strlen(ref) + 16];
  strcpy(path, lazy->path);
  strcat(path, ref);
  char *cp = strrchr(path, '\n');
  if (cp) *cp = '\0';

  char *slash;
  cp = path + lazy->pathlen;
  while ((slash = strchr(cp, '/'))) {
    *slash = '\0';
    if (!direxists(path)) bmkdir(path);
    *slash = '/';
    cp = slash + 1;
  }
  int fd = bopen_write(path);
  write(fd, buf, len);  // TODO: Write bwrite().
  close(fd);
}

static int lazy_has_object(db_ptr db, const ID id) {
  die("has_object() unimplemented. Do you really need it?"
      " It is expensive for lazy repos.");
}

static void lazy_forall_branches(db_ptr db, void (*fn)(const char *ref)) {
  char s[PATH_MAX];
  int slen = 0, oldlen;
  s[0] = '\0';
  void walk(const char *path) {
    int bmode = blstat(path);
    switch(bmode) {
    case bmode_dir:
      oldlen = slen;
      if (slen) strcat(s, "/");
      strcat(s, path);
      slen = strlen(s);
      bchdir(path);
      bopendir(".", walk);
      bchdir("..");
      slen = oldlen;
      s[slen] = '\0';
      break;
    case bmode_reg:
      if (slen) strcat(s, "/");
      strcat(s, path);
      fn(s);
      s[slen] = '\0';
      break;
    default:
      fprintf(stderr, "ignoring bad branch %s", path);
      break;
    }
  }
  lazydb_ptr lazy = db_data(db);
  char ref_dir[lazy->pathlen + 16];
  sprintf(ref_dir, "%srefs/heads/", lazy->path);
  bchdir(ref_dir);
  bopendir(".", walk);
}

static void lazy_appendlog(db_ptr db, const ID id, const char *msg) {
  lazydb_ptr lazy = db_data(db);
  char logfile[lazy->pathlen + 8];
  sprintf(logfile, "%slog", lazy->path);
  FILE *fp = bfopen(logfile, "ab");
  id_fprint(fp, id);
  fprintf(fp, " %s\n", msg);
  bfclose(fp);
}

// db_new_lazy() should be the only extern function.
db_ptr db_new_lazy(const char *path, const char *url) {
  db_ptr db = malloc(sizeof(*db));
  db->data = malloc(sizeof(struct lazydb_s));
  lazydb_ptr lazy = db_data(db);
  lazy->path = strdup(path);
  lazy->pathlen = strlen(path);

  lazy->url = strdup(url);
  lazy->urllen = strlen(url);
  lazy->tmpdir = malloc(16);
  strcpy(lazy->tmpdir, "/tmp/GG_TMP_XXXXXX");
  if (!mkdtemp(lazy->tmpdir)) die_perror("mkdtemp");
  lazy->tmplen = strlen(lazy->tmpdir);

  db->delme = lazy_delete;
  db->reset = lazy_reset;
  db->write_ref_raw = lazy_write_ref_raw;
  db->mmap_ref_raw = lazy_mmap_ref_raw;
  db->has_branch = lazy_has_branch;
  db->forall_branches = lazy_forall_branches;
  db->has_object = lazy_has_object;
  db->mmap_object_raw = lazy_mmap_object_raw;
  db->write_header_body_raw = lazy_write_header_body_raw;
  db->appendlog = lazy_appendlog;
  return db;
}
