#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <unistd.h>
#include "bstd.h"
#include "id.h"
#include "db_t.h"

// URL object database: all information stored remotely.
// Limited to read-only operations.

struct urldb_s {
  char *path;
  int pathlen;
  char *tmpdir;
  int tmplen;
};
typedef struct urldb_s *urldb_ptr;

static void *mmap_download(size_t *len, urldb_ptr u, const char *url) {
  char cmd[strlen(url) + 128];
  ID id;
  char file[u->tmplen + ID_SLEN + 2];
  id_string(id, url);
  sprintf(file, "%s/", u->tmpdir);
  id_sprint(file + u->tmplen + 1, id);
  sprintf(cmd, "curl %s -o %s", url, file);
  printf("command: %s\n", cmd);
  if (system(cmd)) die("system() error: %s", cmd);
  void *res = map_file_to_memory(len, file);
  // TODO: Cache file instead.
  unlink(file);
  return res;
}

static void *url_mmap_object_raw(size_t *len, db_ptr db, const ID id) {
  urldb_ptr u = db_data(db);
  char obj[u->pathlen + PATH_ID_LEN];
  cat_path_id(obj, u->path, id);
  void *res = mmap_download(len, u, obj);
  if (!res) die("invalid object");
  return res;
}

static int url_has_branch(db_ptr db, const char *branch) {
  char needle[strlen(branch) + 16];
  sprintf(needle, "refs/heads/%s", branch);
  int found = 0;
  urldb_ptr u = db_data(db);
  char s[u->pathlen + 32];
  sprintf(s, "%sinfo/refs", u->path);
  size_t len;
  char *buf = mmap_download(&len, u, s);
  char *cp = buf, *lf;
  while ((lf = strchr(cp, '\n'))) {
    char s[lf - cp + 1];
    strncpy(s, cp, lf - cp);
    s[lf - cp] = '\0';
    if (!strcmp(s, needle)) {
      found = 1;
      goto cleanup;
    }
    cp = lf + 1;
    if (cp - buf >= len) break;
  }
cleanup:
  db_munmap(buf, len);
  return found;
}

static void *url_mmap_ref_raw(size_t *len, db_ptr db, const char *ref) {
  urldb_ptr u = db_data(db);
  char s[u->pathlen + strlen(ref) + 1];
  strncpy(s, u->path, u->pathlen);
  strcpy(s + u->pathlen, ref);
  char *cp = strrchr(s, '\n');
  if (cp) *cp = '\0';
  return mmap_download(len, u, s);
}

static void url_write_ref_raw(db_ptr db, const char *ref,
    void *buf, size_t len) {
  die("Can't write ref to read-only repo.");
}

static int url_has_object(db_ptr db, const ID id) {
  die("has_object() unimplemented. Do you really need it?"
      " It is expensive for URL repos.");
}

static void url_forall_branches(db_ptr db, void (*fn)(const char *ref)) {
  urldb_ptr u = db_data(db);
  char s[u->pathlen + 32];
  sprintf(s, "%sinfo/refs", u->path);
  size_t len;
  char *buf = mmap_download(&len, u, s);
  char *cp = buf, *lf;
  while ((lf = strchr(cp, '\n'))) {
    char s[lf - cp + 1];
    strncpy(s, cp, lf - cp);
    s[lf - cp] = '\0';
    printf("branch '%s'\n", s);
    fn(s + strlen("refs/heads/"));
    cp = lf + 1;
    if (cp - buf >= len) break;
  }
  db_munmap(buf, len);
}

static void url_write_header_body_raw(db_ptr db, const ID id,
    const char *header, const unsigned char *buf, size_t len) {
  die("Cannot write to read-only repo.");
}

static void url_appendlog(db_ptr db, const ID id, const char *msg) {
  die("Cannot log to read-only repo.");
}

static void url_delete(db_ptr db) {
  urldb_ptr u = db_data(db);
  free(u->path);
  rmdir(u->tmpdir);
  free(u->tmpdir);
  free(db);
}

static int url_reset(db_ptr db) {
  die("Cannot reset read-only repo.");
}

// db_new_url() should be the only extern function.
db_ptr db_new_url(const char *path) {
  db_ptr db = malloc(sizeof(*db));
  db->data = malloc(sizeof(struct urldb_s));
  urldb_ptr u = db_data(db);
  u->path = strdup(path);
  u->pathlen = strlen(path);
  u->tmpdir = malloc(16);
  strcpy(u->tmpdir, "/tmp/GG_TMP_XXXXXX");
  if (!mkdtemp(u->tmpdir)) die_perror("mkdtemp");
  u->tmplen = strlen(u->tmpdir);

  db->delme = url_delete;
  db->reset = url_reset;
  db->write_ref_raw = url_write_ref_raw;
  db->mmap_ref_raw = url_mmap_ref_raw;
  db->has_branch = url_has_branch;
  db->forall_branches = url_forall_branches;
  db->has_object = url_has_object;
  db->mmap_object_raw = url_mmap_object_raw;
  db->write_header_body_raw = url_write_header_body_raw;
  db->appendlog = url_appendlog;
  return db;
}
