#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include "bstd.h"
#include "id.h"
#include "db_t.h"

#define NDEBUG
#include <assert.h>

inline void db_delete(db_ptr db) {
  db->delme(db);
}

inline int db_reset(db_ptr db) {
  return db->reset(db);
}

inline void *db_mmap_object_raw(size_t *len, db_ptr db, const ID id) {
  return db->mmap_object_raw(len, db, id);
}

inline void db_munmap(void *buf, size_t len) {
  assert(buf);
  munmap(buf, len);
}

inline void *db_mmap_ref_raw(size_t *len, db_ptr db, const char *ref) {
  return db->mmap_ref_raw(len, db, ref);
}

inline void *db_mmap_head_raw(size_t *len, db_ptr db) {
  return db_mmap_ref_raw(len, db, "HEAD");
}

inline void db_read_ref(ID id, db_ptr db, const char *s) {
  size_t len;
  void *buf = db_mmap_ref_raw(&len, db, s);
  if (len < ID_SLEN) die("bad ref: %s", s);
  id_sread(id, buf);
  db_munmap(buf, len);
}

void db_read_head(ID id, db_ptr db) {
  size_t len;
  char *buf = db_mmap_head_raw(&len, db);
  if (len < 6) die("bad HEAD");
  if (!strncmp(buf, "ref: ", 5)) db_read_ref(id, db, buf + 5);
  else {
    if (len < ID_SLEN) die("bad HEAD");
    id_sread(id, buf);
  }
  unmap_memory(buf, len);
}

inline int db_has_branch(db_ptr db, const char *s) {
  return db->has_branch(db, s);
}

void db_read_branch(ID id, db_ptr db, const char *s) {
  char path[strlen(s) + 12];
  sprintf(path, "refs/heads/%s", s);
  db_read_ref(id, db, path);
}

inline void *db_data(db_ptr db) {
  return db->data;
}

inline void db_write_ref_raw(db_ptr db, const char *ref,
    void *buf, size_t len) {
  db->write_ref_raw(db, ref, buf, len);
}

inline void db_write_head_raw(db_ptr db, unsigned char *buf, size_t len) {
  db->write_ref_raw(db, "HEAD", buf, len);
}

int db_write_ref(db_ptr db, const char *ref, const ID id) {
  char buf[ID_SLEN + 1];
  id_sprint(buf, id);
  buf[ID_SLEN] = '\n';
  db_write_ref_raw(db, ref, buf, ID_SLEN + 1);
  return 0;
}

int db_write_branch(db_ptr db, const char *branch, const ID id) {
  char path[strlen(branch) + 16];
  strcpy(path, "refs/heads/");
  strcat(path, branch);
  db_write_ref(db, path, id);
  return 0;
}

int db_seek_branch(db_ptr db, const char *branch) {
  char s[strlen(branch) + 32];
  sprintf(s, "ref: refs/heads/%s\n", branch);
  db_write_head_raw(db, (unsigned char *) s, strlen(s));
  return 0;
}

void db_write_head(db_ptr db, const ID id) {
  // Determine if HEAD is pointing to a branch.
  size_t len;
  char *buf = db_mmap_head_raw(&len, db);
  if (len < 6) die("bad HEAD");
  if (!strncmp(buf, "ref: ", 5)) db_write_ref(db, buf + 5, id);
  else {
    char s[ID_SLEN + 2];
    id_sprint(s, id);
    s[ID_SLEN] = '\n';
    s[ID_SLEN + 1] = '\0';
    db_write_head_raw(db, (unsigned char *) s, ID_SLEN + 2);
  }
  unmap_memory(buf, len);
}

int db_stat(db_ptr db, const ID id, size_t *len) {
  size_t buflen;
  char *buf = db_mmap_object_raw(&buflen, db, id);
  if (!buflen) return -1;
  int type = obj_type(*buf);

  if (type < 0 || buflen < 2) {
    db_munmap(buf, buflen);
    return -1;
  }

  switch(buf[1]) {
  case 'r':  // Raw format: stat file to get size.
    // Subtract 1 for header, 1 for type.
    *len = buflen - 2;
    break;
  case 'z':  // zlib: next 2+ bytes hold the size.
    {
    int i = 0;
    if (i + 2 >= buflen) die("bad blob: unexpected EOF");
    *len = (unsigned char) buf[i + 2];
    unsigned char c;
    do {
      i++;
      if (i + 2 >= buflen) die("bad blob: unexpected EOF");
      c = buf[i + 2];
      *len += c << (i * 8);
    } while (c & (1 << 7));
    }
    break;
  default:
    fprintf(stderr, "id: ");
    id_fprint(stderr, id);
    fprintf(stderr, "\n");
    die("unrecognized body format '%c'", buf[1]);
      break;
  }

  db_munmap(buf, buflen);
  return type;
}

inline void db_forall_branches(db_ptr db, void (*fn)(const char *ref)) {
  db->forall_branches(db, fn);
}

#include <zlib.h>

// All objects use the same body format.
//
// Body format: <tag> <data>
//
// Format of <data> depends on <tag>:
//   'r' <raw data>
//   'z' <size> <zlib compressed data>
//
// <size> := <lowest byte> <higher byte>+
//
// We stop reading higher bytes once its highest bit is clear.
// e.g. two bytes hold file sizes from 0 to 7fff.
// ffff is represented by ff ff 00, and so on.
// The size is that of the uncompressed data.
//
// Combining the format specifiier with the 'b' would save 1 byte per blob
// (my original plan), but easy blob detection is handy.

enum {
  format_raw = 0,
  format_zlib,
};
static int preferred_format = format_zlib;

void *encode_body(size_t *outlen, const unsigned char *buf, size_t len) {
  unsigned char *outbuf;
  switch(preferred_format) {
  case format_raw:
    *outlen = len + 1;
    outbuf = malloc(*outlen);
    *outbuf = 'r';
    memcpy(outbuf + 1, buf, len);
    break;
  case format_zlib:
    {
    uLong zlen = compressBound(len);
    outbuf = malloc(zlen + 32);
    *outbuf = 'z';
    size_t sz = len;
    int i = 1;
    outbuf[i] = sz;
    do {
      sz >>= 8;
      i++;  // This shouldn't exceed 32!
      outbuf[i] = sz;
    } while (sz >= (1 << 7));
    i++;
    int status = compress(outbuf + i, &zlen, buf, len);
    *outlen = zlen;
    if (Z_OK != status) {
      die("zlib: compression error %d", status);
    }
    }
    break;
  }
  return outbuf;
}

inline int db_has_object(db_ptr db, const ID id) {
  return db->has_object(db, id);
}

void *map_body_to_memory(size_t *outlen, const unsigned char *buf, size_t len) {
  void *outbuf;
  switch (*buf) {
  case 'r':
    *outlen = len - 1;
    outbuf = mmap_alloc(len - 1);
    memcpy(outbuf, buf + 1, len - 1);
    break;
  case 'z':
    ;
    *outlen = buf[1];
    int i = 2;
    unsigned char c;
    do {
      c = buf[i];
      *outlen += ((uLongf) c) << ((i - 1) * 8);
      i++;
    } while (c & (1 << 7));
    outbuf = mmap_alloc(*outlen);
    uLongf sz = *outlen;
    uncompress(outbuf, &sz, buf + i, len - i);
    if (sz != *outlen) die("uncompress: length mismatch");
    break;
  default:
    die("unhandled data format");
    outbuf = NULL;
  }
  return outbuf;
}

void *db_mmap_body(size_t *bodylen, int *type, db_ptr db, const ID id) {
  size_t len;
  char *buf = db_mmap_object_raw(&len, db, id);
  if (!len) die("invalid object");
  if (obj_type(*buf) < 0) die("invalid object type");
  if (type) *type = obj_type(*buf);
  void *bodybuf = map_body_to_memory(bodylen,
      (unsigned char *) buf + 1, len - 1);
  db_munmap(buf, len);
  return bodybuf;
}

void *db_mmap_body_check_type(size_t *bodylen, int type,
    db_ptr db, const ID id) {
  size_t len;
  char *buf = db_mmap_object_raw(&len, db, id);
  if (!len) die("invalid object");
  if (obj_type(*buf) != type) die("object type mismatch");
  void *bodybuf = map_body_to_memory(bodylen,
      (unsigned char *) buf + 1, len - 1);
  db_munmap(buf, len);
  return bodybuf;
}

inline void db_write_header_body_raw(db_ptr db, const ID id,
    const char *header, const unsigned char *buf, size_t len) {
  db->write_header_body_raw(db, id, header, buf, len);
}

inline void db_write_object_raw(db_ptr db, const ID id,
    unsigned char *buf, size_t len) {
  db_write_header_body_raw(db, id, "", buf, len);
}

// TODO: Add version that computes hash as well?
void db_write_object(db_ptr db, const ID id,
    int type, unsigned char *buf, size_t len) {
  const char *header = typename(type);

  size_t enclen;
  unsigned char *encbuf = encode_body(&enclen, buf, len);
  db_write_header_body_raw(db, id, header, encbuf, enclen);
  free(encbuf);
}

inline void db_log(db_ptr db, const ID id, const char *msg) {
  db->appendlog(db, id, msg);
}
