#define __DB_H__
// requires <stdio.h>
// requires "id.h"

// Instead of just abstracting transports, we provide an opaque interface to
// the object database. For all the caller knows, the database could be an
// HTTP server, a local directory, a local directory but with everything in
// nonstandard locations, and so on.

struct db_s;
typedef struct db_s *db_ptr;

void *db_data(db_ptr db);

struct index_s;
typedef struct index_s *index_ptr;

// Return a classic filesystem-based database.
db_ptr db_new_gg(const char *path);
// Return a URL-based database.
db_ptr db_new_url(const char *url);
// Return a hybrid database.
db_ptr db_new_lazy(const char *path, const char* url);

// Call when finished with db.
void db_delete(db_ptr db);

// Create or reinitialize database.
int db_reset(db_ptr db);

void *db_mmap_ref_raw(size_t *len, db_ptr db, const char *ref);
void db_write_ref_raw(db_ptr db, const char *ref, void *buf, size_t len);
// Read/write hash to given ref.
void db_read_ref(ID id, db_ptr db, const char *s);
int db_write_ref(db_ptr db, const char *ref, const ID id);
// Read/write hash to given branch. Same as above except
// prepends "refs/heads/" to given branch name.
void db_read_branch(ID id, db_ptr db, const char *branch);
int db_write_branch(db_ptr db, const char *branch, const ID id);
int db_has_branch(db_ptr db, const char *s);
void db_forall_branches(db_ptr db, void (*fn)(const char *ref));

// Read/write ID that HEAD points to, following ref if necessary.
void db_read_head(ID id, db_ptr db);
void db_write_head(db_ptr db, const ID id);
// Read/write raw data to HEAD. Do not follow refs.
void *db_mmap_head_raw(size_t *len, db_ptr db);
void db_write_head_raw(db_ptr db, unsigned char *buf, size_t len);
// Move HEAD to given branch.
int db_seek_branch(db_ptr db, const char *branch);

int db_has_object(db_ptr db, const ID id);
// Map contents of raw object to memory.
void *db_mmap_object_raw(size_t *len, db_ptr db, const ID id);
// Same, but also decompress object.
void *db_mmap_body(size_t *len, int *type, db_ptr db, const ID id);
// Same, but only if object type matches.
void *db_mmap_body_check_type(size_t *len, int type, db_ptr db, const ID id);
// Free memory taken by mmap.
void db_munmap(void *buf, size_t len);
int db_stat(db_ptr db, const ID id, size_t *len);
void db_write_object(db_ptr db, const ID id,
    int type, unsigned char *buf, size_t len);
void db_write_object_raw(db_ptr db, const ID id,
    unsigned char *buf, size_t len);


void db_read_index(index_ptr index, db_ptr db);
void db_write_index(db_ptr db, index_ptr index);

void db_log(db_ptr db, const ID id, const char *msg);

void parse_rev(ID id, db_ptr db, const char *s);

enum object_type {
  OBJ_BLOB = 0,
  OBJ_TREE,
  OBJ_COMMIT,
  OBJ_TAG,
};
const char *typename(int type);

static inline int obj_type(char c) {
  switch (c) {
  case 'b':
    return OBJ_BLOB;
  case 't':
    return OBJ_TREE;
  case 'c':
    return OBJ_COMMIT;
  case 'g':
    return OBJ_TAG;
  default:
    return -1;
  }
}

size_t forall_executables(db_ptr db, const ID mode_id,
    void (*callback)(const char *filename));

enum {
  // strlen("objects/<first byte>/<other bytes>") and terminating NULL.
  PATH_ID_LEN = ID_SLEN + 10,
};

// 's' must have enough room for the path + PATH_ID_LEN bytes.
static inline void cat_path_id(char *s, const char *path, const ID id) {
  sprintf(s, "%sobjects/%02x/", path, id[0]);
  char buf[ID_SLEN + 1];
  id_sprint(buf, id);
  strcat(s, buf + 2);
}
