// Ben Lynn
// start: Thu May 29 20:36:53 PDT 2008
#define _GNU_SOURCE
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include "bstd.h"
#include "id.h"
#include "darray.h"
#include "cbt.h"
#include "db.h"
#include "fstree.h"
#include "index.h"
#include "tree.h"
#include "commit.h"
#include "list.h"

#define NDEBUG
#include <assert.h>

void usage() {
  printf("Usage: gg COMMAND\n");
  exit(0);
}

// Create the eponymous master branch with zero commit
// and a HEAD reference pointing to the master.
// TODO: Do not zero the master branch if it already exists.
int gg_init_db(int argc, const char **argv, db_ptr db) {
  db_reset(db);
  ID id;
  id_set0(id);
  db_write_branch(db, "master", id);
  db_seek_branch(db, "master");
  return 0;
}

const char *typename(int type) {
  static char *names[] = { "b", "t", "c", "g" };
  if (type < 0 || type > 3) die("no such type %d", type);
  return names[type];
}

static void blob_fwrite(db_ptr db, const ID id, FILE *dst) {
  size_t len;
  unsigned char *buf = db_mmap_body_check_type(&len, OBJ_BLOB, db, id);
  fwrite(buf, 1, len, dst);
  db_munmap(buf, len);
}

int gg_add(int argc, const char **argv, db_ptr db) {
  if (argc < 2) {
    printf("Nothing ventured, nothing gained.\n");
    return 0;
  }
  index_t index;
  db_index_init(index, db);
  index_read(index);
  int i;
  for (i = 1; i < argc; i++) {
    index_add_path(index, db, argv[i]);
  }
  index_write(index);
  index_clear(index);
  return 0;
}

// TODO: With more function pointers, write directly from fstree to file.
void write_tree_from_fstree(db_ptr db, ID id,
    fstree_t f, unsigned char *(*fn)(void *)) {
  tree_t tree;
  tree_init(tree);
  void add_child(void *p, const char *key) {
    ID childid;
    fstree_ptr fchild = p;
    write_tree_from_fstree(db, childid, fchild, fn);
    tree_append_tree(tree, childid, key);
  }
  void add_entry(void *p, const char *key) {
    tree_append_blob(tree, fn(p), key);
  }
  cbt_forall(f->child, add_child);
  cbt_forall(f->entry, add_entry);
  compute_id_and_write_tree(db, id, tree);
  tree_clear(tree);
}

// Write all the names out in a list, with a header.
//
// One argument is a pointer to a function taking a pointer to a function:
// the supplied function iterates over all filenames, and runs a given callback
// on each filename.
void write_mode_blob_helper(db_ptr db, ID id,
    void (*forall_fn)(void (*)(void *)), size_t size) {
  char *s = malloc(size + 1);  // 1 extra for handy NULL terminator.
  // Write format tag: 0 means list of executables delimited by newlines.
  s[0] = '0';
  s[1] = '\n';
  char *cp = s + 2;
  void cat(void *p) {
    int k = strlen(p);
    memcpy(cp, p, k);
    cp += k;
    *cp = '\n';
    cp++;
  }
  // list_forall(xfiles, cat);
  forall_fn(cat);
  *cp = '\0';
  assert(size - 1 == strlen(s));

  // Hash and write to blob object. I've considered defining a new object
  // type, but don't see the point.
  id_header_string(id, "b", s);
  db_write_object(db, id, OBJ_BLOB, (unsigned char *) s, strlen(s));
  free(s);
}

// Write a blob containing mode and permissions information.
// TODO: Move to index.c. Better yet, do this during write_tree.
#include <sys/types.h>
#include <sys/stat.h>
static void write_mode_blob(db_ptr db, ID id, fstree_t f) {
  // 1 for format tag, 1 for newline after.
  size_t size = 2;

  // Go through whole index looking for executables.
  // When we find one, copy its full pathname and add its
  // size to a running total.
  list_t dirstack;  // For building up full pathnames.
  list_t xfiles;  // List of executable files.
  list_init(dirstack);
  list_init(xfiles);
  void append(fstree_ptr f) {
    void check_child(void *p, const char *key) {
      fstree_ptr fchild = p;
      list_append(dirstack, (void *) key);
      append(fchild);
      list_remove_last(dirstack);
    }
    void check_entry(void *p, const char *key) {
      index_entry_ptr ie = p;
      if (S_IXUSR & ie->mode) {
	int n = strlen(key) + 1;
	void addn(void *p) { n += strlen(p) + 1; }
	list_forall(dirstack, addn);
	char *fullname = malloc(n);
	char *cp = fullname;

	void cat(void *p) {
	  int k = strlen(p);
	  memcpy(cp, p, k);
	  cp[k] = '/';
	  cp += k + 1;
	}
	list_forall(dirstack, cat);
	strcpy(cp, key);
	assert(strlen(fullname) + 1 == n);
	list_append(xfiles, fullname);
	size += n;
      }
    }
    cbt_forall(f->child, check_child);
    cbt_forall(f->entry, check_entry);
  }
  append(f);
  list_clear(dirstack);
  void forallx(void (*fn)(void *)) {
    list_forall(xfiles, fn);
  }
  write_mode_blob_helper(db, id, forallx, size);
  list_clear_with(xfiles, free);
}

// Like git write-tree, but also makes a blob describing file
// modes and permissions.
void objectify_index(db_ptr db, ID id, ID xid, index_t index) {
  write_mode_blob(db, xid, index->root);
  unsigned char *return_id(void *p) {
    index_entry_ptr e = p;
    return e->id;
  }
  write_tree_from_fstree(db, id, index->root, return_id);
}

void launch_editor(const char *path);

// From Git's date.c.
static int date_string(unsigned long date, int offset, char *buf, int len) {
  int sign = '+';

  if (offset < 0) {
    offset = -offset;
    sign = '-';
  }
  return snprintf(buf, len, "%lu %c%02d%02d",
      date, sign, offset / 60, offset % 60);
}

// Need this to figure out local timezone offset.
static time_t my_mktime(struct tm *tm) {
  static const int mdays[] = {
    0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334
  };
  int year = tm->tm_year - 70;
  int month = tm->tm_mon;
  int day = tm->tm_mday;

  if (year < 0 || year > 129) /* algo only works for 1970-2099 */
	  return -1;
  if (month < 0 || month > 11) /* array bounds */
	  return -1;
  if (month < 2 || (year + 2) % 4)
	  day--;
  return (year * 365 + (year + 1) / 4 + mdays[month] + day) * 24*60*60UL +
	  tm->tm_hour * 60*60 + tm->tm_min * 60 + tm->tm_sec;
}

void datestamp(char *buf, int bufsize) {
  time_t now;
  int offset;

  time(&now);
  // Figure out local timezone offset through a hack.
  offset = my_mktime(localtime(&now)) - now;
  offset /= 60;
  date_string(now, offset, buf, bufsize);
}

const char *gg_committer_id() {
  char *s = getenv("GG_COMMITTER_ID");
  if (!s) {
    die("Please set GG_COMMITTER_ID to 'Your Name <your_email@example.com>'");
  }
  return s;
}

int gg_commit(int argc, const char **argv, db_ptr db) {
  index_t index;
  db_index_init(index, db);
  index_read(index);

  commit_t commit;
  commit_init(commit);

  // We may be working with a remote gg repository.
  // The message should be created on local disk, so we avoid the .gg
  // directory (it might not even exist).
  // I'd like to use tmpfile(), but I don't know how to get its filename,
  // which must be given to the launched editor.
  char commitfile[80];
  strcpy(commitfile, "/tmp/COMMIT_EDITMSG_XXXXXX");
  int fd = mkstemp(commitfile);
  if (!fd) die_perror("mkstemp %s", commitfile);
  close(fd);
  launch_editor(commitfile);
  commit->message = bfiletostr(&commit->messagelen, commitfile);
  if (!commit->messagelen) die("Empty commit message");

  objectify_index(db, commit->tree_id, commit->mode_id, index);

  index_clear(index);

  ID id;
  db_read_head(id, db);

  if (!id_is0(id)) commit_add_parent(commit, id);

  commit->committer = malloc(1337);
  char dates[64];
  datestamp(dates, 64);
  snprintf(commit->committer, 1337, "%s %s",
      gg_committer_id(), dates);
  commit_write(db, id, commit);
  db_write_head(db, id);
  db_log(db, id, "commit");
  commit_clear(commit);
  bunlink(commitfile);

  return 0;
}

static void index_from_tree(fstree_ptr f, tree_ptr t, db_ptr db) {
  void read_tree(void *p) {
    entry_ptr e = p;
    tree_t tree;
    tree_init(tree);
    tree_read(tree, db, e->id);
    fstree_ptr fchild = malloc(sizeof(fstree_t));
    fstree_init(fchild);
    assert(!cbt_at(f->entry, e->s));  // e->s cannot also be filename.
    index_from_tree(fchild, tree, db);
    tree_clear(tree);
    cbt_put(f->child, fchild, e->s);
  }
  void read_blob(void *p) {
    entry_ptr e = p;
    index_entry_ptr ie = malloc(sizeof(index_entry_t));
    id_cpy(ie->id, e->id);
    assert(!cbt_at(f->child, e->s));  // e->s cannot also be dirname.
    cbt_put(f->entry, ie, e->s);
  }
  darray_forall(t->tree_list, read_tree);
  darray_forall(t->blob_list, read_blob);
}

// TODO: Handle case where same name but different type already exists.
void fstree_checkout(fstree_ptr f, db_ptr db) {
  void write_file(void *p, const char *key) {
    index_entry_ptr ie = p;
    FILE *dst = bfopen(key, "wb");
    blob_fwrite(db, ie->id, dst);
    bfclose(dst);
  }
  void write_dir(void *p, const char *key) {
    fstree_ptr fchild = p;
    if (!direxists(key)) {
      bmkdir(key);
    }
    bchdir(key);
    fstree_checkout(fchild, db);
    bchdir("..");
  }
  cbt_forall(f->entry, write_file);
  cbt_forall(f->child, write_dir);
}

// Read list of executables and call supplied callback.
// Return total size of mode blob.
size_t forall_executables(db_ptr db, const ID mode_id,
    void (*callback)(const char *filename)) {
  if (id_is0(mode_id)) return 0;
  size_t len;
  char *buf = db_mmap_body_check_type(&len, OBJ_BLOB, db, mode_id);
  if ('\n' != buf[len - 1]) die("missing newline at end of file");
  char *cp = buf;
  if ('0' != *cp) die("unrecognized mode info format");
  cp++;
  if ('\n' != *cp) die("missing newline after format tag");
  cp++;
  while (cp != buf + len) {
    char *lf = strchr(cp, '\n');
    *lf = '\0';
    callback(cp);
    cp = lf + 1;
  }
  unmap_memory(buf, len);
  return len;
}

void index_checkout(index_t index, db_ptr db, ID mode_id) {
  fstree_checkout(index->root, db);
  forall_executables(db, mode_id, bmake_executable);
}

void parse_rev(ID id, db_ptr db, const char *s) {
  // Translate string into ID, trying the following in order:
  // 1. "HEAD"
  if (!strcmp(s, "HEAD")) {
    db_read_head(id, db);
    return;
  }

  // 2. A branch or tag name.
  if (db_has_branch(db, s)) {
    db_read_branch(id, db, s);
    return;
  }

  // 3. A SHA1 hash referring to some node on the history graph.
  if (ID_SLEN != strlen(s)) die("expect 40-byte hash\n");
  id_sread(id, s);
}

int gg_checkout(int argc, const char **argv, db_ptr db) {
  if (argc < 2) {
    die("Must supply hash of commit");
  }
  ID id;
  // TODO: Distinguish between branches and commits, handle accordingly.
  parse_rev(id, db, argv[1]);
  index_t index;
  db_index_init(index, db);
  db_write_head(db, id);
  db_log(db, id, "checkout");
  if (id_is0(id)) {
    printf("Zero commit: emptying index...\n");
    index_write(index);
    return 0;
  }

  commit_t commit;
  commit_init(commit);
  commit_read(commit, db, id);

  tree_t tree;
  tree_init(tree);
  tree_read(tree, db, commit->tree_id);

  index_from_tree(index->root, tree, db);
  tree_clear(tree);

  index_checkout(index, db, commit->mode_id);
  index_write(index);
  index_clear(index);

  commit_clear(commit);
  
  return 0;
}

int gg_log(int argc, const char **argv, db_ptr);

void export(db_ptr, ID);
void import(db_ptr, int);

int gg_export(int argc, const char **argv, db_ptr db) {
  if (argc < 2) {
    die("Must supply hash of commit");
  }
  if (40 != strlen(argv[1])) {
    die("Commit must be given in hex");
  }
  ID id;
  id_sread(id, argv[1]);
  export(db, id);
  
  return 0;
}

int gg_import(int argc, const char **argv, db_ptr db) {
  import(db, 0);
  
  return 0;
}

// Git uses from commands to mark parents.
// Assuming this prevents fake relationships for git-fast-exports.
// A commit is considered initial unless a from command is seen.
int gg_git_import(int argc, const char **argv, db_ptr db) {
  import(db, 1);
  
  return 0;
}

// icc stands for index_curdir_cmp, a complicated function needing a bunch
// of callbacks:
struct icc_callback_s {
  // Callback called when...
  // TODO: Also pass index_entry_ptr.
  void (*ireg)(list_t, const char *);   // index-only file encountered.
  void (*dreg)(list_t, const char *);   // dir-only file encountered.

  // file present in both index and working directory encountered.
  void (*idreg)(list_t, index_entry_ptr, const char *);

  void (*idir)(list_t);   // index-only subdirectory encountered.
  void (*ddir)(list_t);   // dir-only subdirectory encountered.
  void (*dirin)(const char *);  // entering subdirectory.
  void (*dirout)();  // leaving subdirectory.
};
typedef struct icc_callback_s icc_callback_t[1];
typedef struct icc_callback_s *icc_callback_ptr;

static void stackprint(const char *status,
    list_t stack, const char *pathname) {
  printf("%s: ", status);
  void printdir(void *p) {
    printf("%s/", (char *) p);
  }
  list_forall(stack, printdir);
  printf("%s\n", pathname);
}

// Walk through the current directory, comparing the paths to those in the
// index. Call given callbacks accordingly.
void index_curdir_cmp(index_ptr index, icc_callback_t cb) {
  list_t dirstack;
  list_init(dirstack);
  void index_vs_workdir(fstree_ptr f) {
    cbt_t d_roll;
    cbt_t r_roll;
    cbt_init(d_roll);
    cbt_init(r_roll);
    void do_path(const char *pathname) {
      int bmode = blstat(pathname);
      switch(bmode) {
      case bmode_dir:
	list_append(dirstack, (void *) pathname);
	fstree_ptr fchild = cbt_at(f->child, pathname);
	if (fchild) {
	  cbt_put(d_roll, (void *) 1, pathname);
	  if (cb->dirin) cb->dirin(pathname);
	  bchdir(pathname);
	  index_vs_workdir(fchild);
	  bchdir("..");
	  if (cb->dirout) cb->dirout();
	} else {
	  if (cb->ddir) cb->ddir(dirstack);
	}
	list_remove_last(dirstack);
	break;
      case bmode_reg:
      {
	index_entry_ptr ie = cbt_at(f->entry, pathname);
	if (!ie) {
	  if (cb->dreg) cb->dreg(dirstack, pathname);
	} else {
	  cbt_put(r_roll, (void *) 1, pathname);
	  if (cb->idreg) cb->idreg(dirstack, ie, pathname);
	}
      }
	break;
      // TODO: Handle links.
      default:
	stackprint("???", dirstack, pathname);
      }
    }
    bopendir(".", do_path);

    // Call a callback for everybody missing from the roll.
    void rollcall(cbt_t roll, cbt_t everybody, void (*fn)(const char *)) {
      cbt_it roll_it = cbt_it_first(roll);
      cbt_it it = cbt_it_first(everybody);

      while(!cbt_it_is_off(roll_it)) {
	while (strcmp(cbt_it_key(roll_it), cbt_it_key(it))) {
	  assert(0 < strcmp(cbt_it_key(roll_it), cbt_it_key(it)));
	  fn(cbt_it_key(it));
	  it = cbt_it_next(it);
	}
	roll_it = cbt_it_next(roll_it);
	it = cbt_it_next(it);
      }
      while(!cbt_it_is_off(it)) {
	fn(cbt_it_key(it));
	it = cbt_it_next(it);
      }
    }
    // Look for files in index absent from working directory.
    void ireg(const char *name) {
      if (cb->ireg) cb->ireg(dirstack, name);
    }
    rollcall(r_roll, f->entry, ireg);
    cbt_clear(r_roll);

    // Look for subdirectories in index absent from working directory.
    void idir(const char *name) {
      if (cb->idir) {
	list_append(dirstack, (void *) name);
	cb->idir(dirstack);
	list_remove_last(dirstack);
      }
    }
    rollcall(d_roll, f->child, idir);
    cbt_clear(d_roll);
  }
  index_vs_workdir(index->root);
}

static void index_curdir_status(index_t index) {
  void ireg(list_t stack, const char *filename) {
    stackprint("deleted", stack, filename);
  }
  void dreg(list_t stack, const char *filename) {
    stackprint("untracked", stack, filename);
  }
  void both(list_t stack, index_entry_ptr ie, const char *filename) {
    if (index_entry_cmp(index, ie, filename)) {
      stackprint("modified", stack, filename);
    }
  }
  void idir(list_t stack) {
    stackprint("deleted", stack,  "");
  }
  void ddir(list_t stack) {
    stackprint("untracked", stack,  "");
  }
  icc_callback_t cb = {{ ireg, dreg, both, idir, ddir, NULL, NULL }};
  index_curdir_cmp(index, cb);
}

// Like commit -a
// i.e. find differences between working directory and HEAD among
// but ignoring files present neither in the HEAD or index.
int gg_save(int argc, const char **argv, db_ptr db) {
  index_t index, new_index;
  db_index_init(index, db);
  index_read(index);
  db_index_init(new_index, db);

  fstree_ptr fnew = new_index->root;
  list_t fstack;
  list_init(fstack);
  fstree_ptr fnow = fnew;

  void ireg(list_t stack, const char *filename) {
    stackprint("deleted", stack, filename);
  }
  void dreg(list_t stack, const char *filename) {
    stackprint("untracked", stack, filename);
  }
  void both(list_t stack, index_entry_ptr ie, const char *filename) {
    ie = index_entry_dup(ie);
    if (fstree_update_path(index, db, fnow, ie, filename)) {
      stackprint("modified", stack, filename);
    }
  }
  void idir(list_t stack) {
    stackprint("deleted", stack, "");
  }
  void ddir(list_t stack) {
    stackprint("untracked", stack, "");
  }
  void in(const char *dirname) {
    fstree_ptr fchild = malloc(sizeof(*fchild));
    fstree_init(fchild);
    list_append(fstack, fnow);
    cbt_put(fnow->child, fchild, dirname);
    fnow = fchild;
  }
  void out() {
    fnow = list_remove_last(fstack);
  }
  icc_callback_t cb = {{ ireg, dreg, both, idir, ddir, in, out }};
  index_curdir_cmp(index, cb);
  list_clear(fstack);

  // TODO: Commit new index.
  printf("Please type 'gg commit'\n");
  index_clear(index);

  index_write(new_index);
  index_clear(new_index);
  return 0;
}

int gg_type(int argc, const char **argv, db_ptr db) {
  if (argc < 2) die("Must supply hash of object");
  if (40 != strlen(argv[1])) die("Hash must be given in hex");
  ID id;
  id_sread(id, argv[1]);
  size_t len;
  char *buf = db_mmap_object_raw(&len, db, id);
  if (!len) die("blob is empty");
  if (obj_type(*buf) < 0) die("unknown object type");
  putchar(*buf);
  putchar('\n');
  db_munmap(buf, len);
  return 0;
}

int gg_cat(int argc, const char **argv, db_ptr db) {
  if (argc < 2) die("Must supply hash of commit");
  if (40 != strlen(argv[1])) die("Commit must be given in hex");
  ID id;
  id_sread(id, argv[1]);

  size_t len;
  unsigned char *buf = db_mmap_body(&len, NULL, db, id);

  fwrite(buf, 1, len, stdout);
  db_munmap(buf, len);
  return 0;
}

int gg_status(int argc, const char **argv, db_ptr db) {
  // First compare head with index.
  ID head_id;
  tree_t head_tree;
  index_t index;
  db_index_init(index, db);
  tree_init(head_tree);
  index_read(index);
  db_read_head(head_id, db);
  if (!id_is0(head_id)) {
    commit_t commit;
    commit_init(commit);
    commit_read(commit, db, head_id);
    tree_read(head_tree, db, commit->tree_id);
    commit_clear(commit);
  }

  // fstree stores lists as crit-bit trees, and tree as darrays.
  // We exploit that the lists are sorted.
  printf("HEAD vs index:\n");
  void file_cmp(cbt_t flist, darray_t tlist) {
    cbt_it index_it = cbt_it_first(flist);
    int head_i = 0;
    for (;;) {
      if (head_i == darray_count(tlist)) {
	while (!cbt_it_is_off(index_it)) {
	  printf("adding: %s\n", cbt_it_key(index_it));
	  index_it = cbt_it_next(index_it);
	}
	break;
      }
      if (cbt_it_is_off(index_it)) {
	while (head_i < darray_count(tlist)) {
	  entry_ptr e = darray_at(tlist, head_i);
	  printf("deleting: %s\n", e->s);
	  head_i++;
	}
	break;
      }
      entry_ptr e = darray_at(tlist, head_i);
      char *index_s = cbt_it_key(index_it);
      int res = strcmp(e->s, index_s);
      if (res < 0) {
	printf("deleting: %s\n", e->s);
	head_i++;
      } else if (res > 0) {
	printf("adding: %s\n", index_s);
	index_it = cbt_it_next(index_it);
      } else {
	index_entry_ptr ie = cbt_it_data(index_it);
	if (id_cmp(e->id, ie->id)) {
	  printf("modified: %s\n", index_s);
	}
	head_i++;
	index_it = cbt_it_next(index_it);
      }
    }
  }
  void dir_cmp(fstree_ptr f, tree_ptr t) {
    file_cmp(f->entry, t->blob_list);

    cbt_it index_it = cbt_it_first(f->child);
    int head_i = 0;
    // TODO: Add dirstack.
    for(;;) {
      if (head_i == darray_count(t->tree_list)) {
	while (!cbt_it_is_off(index_it)) {
	  // TODO: show files, not just the dir
	  printf("adding some or all of: %s/\n", cbt_it_key(index_it));
	  index_it = cbt_it_next(index_it);
	}
	break;
      }
      if (cbt_it_is_off(index_it)) {
	while (head_i < darray_count(t->tree_list)) {
	  entry_ptr e = darray_at(t->tree_list, head_i);
	  printf("deleting: %s/\n", e->s);
	  head_i++;
	}
	break;
      }
      entry_ptr e = darray_at(t->tree_list, head_i);
      char *index_s = cbt_it_key(index_it);
      int res = strcmp(e->s, index_s);
      if (res < 0) {
	printf("deleting: %s/\n", e->s);
	head_i++;
      } else if (res > 0) {
	// TODO: show files, not just the dir
	printf("adding some or all of: %s/\n", index_s);
	index_it = cbt_it_next(index_it);
      } else {
	tree_t subtree;
	tree_init(subtree);
	tree_read(subtree, db, e->id);
	dir_cmp(cbt_it_data(index_it), subtree);
	tree_clear(subtree);
	head_i++;
	index_it = cbt_it_next(index_it);
      }
    }
  }
  dir_cmp(index->root, head_tree);

  // Now compare index with working directory.
  puts("index vs working dir:");
  index_curdir_status(index);

  // Write index because we may have cleared some dodgy flags.
  index_write(index);
  index_clear(index);
  tree_clear(head_tree);
  return 0;
}

int gg_diff(int argc, const char **argv, db_ptr db) {
  index_t index;
  void ireg(list_t stack, const char *filename) {
    stackprint("Only in index", stack, filename);
  }
  void dreg(list_t stack, const char *filename) {
    stackprint("Only in working directory", stack, filename);
  }
  void both(list_t stack, index_entry_ptr ie, const char *filename) {
    if (index_entry_cmp(index, ie, filename)) {
      char cmd[1337];
      FILE *fp;
      snprintf(cmd, 1337, "diff -u - %s", filename);
      fp = popen(cmd, "w");
      blob_fwrite(db, ie->id, fp);
      pclose(fp);
    }
  }
  void idir(list_t stack) {
    stackprint("Only in index", stack, "");
  }
  void ddir(list_t stack) {
    stackprint("Only in working directory", stack, "");
  }
  icc_callback_t cb = {{ ireg, dreg, both, idir, ddir, NULL, NULL }};
  if (argc < 2) {
    // Compare index with working directory.
    db_index_init(index, db);
    index_read(index);
    index_curdir_cmp(index, cb);
    index_write(index);
    index_clear(index);
    return 0;
  }
  return 0;
}

void db_copy_object(db_ptr dst, db_ptr src, const ID id) {
  if (!id_is0(id) && !db_has_object(dst, id)) {
    size_t len;
    void *buf = db_mmap_object_raw(&len, src, id);
    db_write_object_raw(dst, id, buf, len);
    unmap_memory(buf, len);
  }
}

static char *guess_dir_name(const char *repo, int is_bundle) {
  const char *p, *start, *end, *limit;
  int after_slash_or_colon;

  /* Guess dir name from repository: strip trailing '/',
   * strip trailing '[:/]*.{git,bundle}', strip leading '.*[/:]'. */

  after_slash_or_colon = 1;
  limit = repo + strlen(repo);
  start = repo;
  end = limit;
  for (p = repo; p < limit; p++) {
    const char *prefix = is_bundle ? ".bundle" : ".gg";
    if (!prefixcmp(p, prefix)) {
      if (!after_slash_or_colon) end = p;
      p += strlen(prefix) - 1;
    } else if (!prefixcmp(p, ".bundle")) {
      if (!after_slash_or_colon) end = p;
      p += 7;
    } else if (*p == '/' || *p == ':') {
      if (end == limit) end = p;
      after_slash_or_colon = 1;
    } else if (after_slash_or_colon) {
      start = p;
      end = limit;
      after_slash_or_colon = 0;
    }
  }

  return strndup(start, end - start);
}

#include "graph.h"
int gg_clone(int argc, const char **argv, db_ptr db) {
  db_delete(db);  // We don't care about current directory in this case.
  if (argc < 2) {
    die("Must supply source directory.");
  }
  const char *srcdir = argv[1];
  int slen = strlen(srcdir);
  char s[PATH_MAX];
  strcpy(s, srcdir);
  if (s[slen - 1] != '/') {
    s[slen] = '/';
    slen++;
    s[slen] ='\0';
  }
  // Set source database.
  db_ptr db_remote;
  if (!strncmp(s, "http:", 5)) {
    db_remote = db_new_url(s);
  } else {
    strcpy(s + slen, "HEAD");
    if (!fileexists(s)) {
      strcpy(s + slen, ".gg/");
      slen += strlen(".gg/");
      strcpy(s + slen, "HEAD");
      if (!fileexists(s)) {
	die("Source not a gg directory");
      }
    }
    s[slen] = '\0';

    char olddir[PATH_MAX];
    if (!getcwd(olddir, PATH_MAX - 100)) die_perror("getcwd");

    bchdir(s);
    if (!getcwd(s, PATH_MAX - 100)) die_perror("getcwd");
    slen = strlen(s);
    s[slen] = '/';
    slen++;
    s[slen] = '\0';
    db_remote = db_new_gg(s);
    bchdir(olddir);
  }

  char *dst = guess_dir_name(s, 0);
  bmkdir(dst);
  bchdir(dst);
  free(dst);
  if (!getcwd(s, PATH_MAX - 100)) die_perror("getcwd");
  strcat(s, "/.gg/");

  // Set target database.
  db = db_new_gg(s);

  db_reset(db);

  ID remote_id;
  db_read_head(remote_id, db_remote);
  graph_t g;
  graph_init(g);
  graph_load_commit(g, db_remote, remote_id, NULL);

  // Copy branches.
  void copy_ref(const char *ref) {
    ID branch_id;
    db_read_branch(branch_id, db_remote, ref);
    graph_load_commit(g, db_remote, branch_id, NULL);
    db_write_branch(db, ref, branch_id);
  }

  db_forall_branches(db_remote, copy_ref);

  // Copy HEAD.
  size_t len;
  void *buf = db_mmap_head_raw(&len, db_remote);
  db_write_head_raw(db, buf, len);
  unmap_memory(buf, len);

  // Copy repo.
  void copy_tree(const ID id) {
    db_copy_object(db, db_remote, id);
    tree_t t;
    tree_init(t);
    tree_read(t, db_remote, id);
    void copy_tree_entry(void *p) {
      entry_ptr e = p;
      copy_tree(e->id);
    }
    darray_forall(t->tree_list, copy_tree_entry);
    void copy_blob_entry(void *p) {
      entry_ptr e = p;
      db_copy_object(db, db_remote, e->id);
    }
    darray_forall(t->blob_list, copy_blob_entry);
    tree_clear(t);
  }
  void copy_commit(commit_ptr c, const ID id) {
    db_copy_object(db, db_remote, id);
    copy_tree(c->tree_id);
    db_copy_object(db, db_remote, c->mode_id);
  }
  graph_forall(g, copy_commit);
  graph_clear(g);
  return 0;
}

int gg_lazy_clone(int argc, const char **argv, db_ptr db) {
  db_delete(db);  // We don't care about current directory in this case.
  if (argc < 2) {
    die("Must supply source directory.");
  }
  const char *srcdir = argv[1];
  int slen = strlen(srcdir);
  char s[PATH_MAX];
  strcpy(s, srcdir);
  if (s[slen - 1] != '/') {
    s[slen] = '/';
    slen++;
    s[slen] ='\0';
  }
  db_ptr db_remote;
  if (!strncmp(s, "http:", 5)) {
    db_remote = db_new_url(s);
  } else {
    die("lazy clone only for urls at the moment");
  }

  char *url = strdup(s);
  char *dst = guess_dir_name(s, 0);
  bmkdir(dst);
  bchdir(dst);
  free(dst);
  if (!getcwd(s, PATH_MAX - 100)) die_perror("getcwd");
  strcat(s, "/.gg/");

  // Set target database.
  db = db_new_lazy(s, url);
  free(url);

  db_reset(db);

  ID remote_id;
  db_read_head(remote_id, db_remote);

  // Copy branches.
  void copy_ref(const char *ref) {
    ID branch_id;
    db_read_branch(branch_id, db_remote, ref);
    db_write_branch(db, ref, branch_id);
  }

  db_forall_branches(db_remote, copy_ref);

  // Copy HEAD.
  size_t len;
  void *buf = db_mmap_head_raw(&len, db_remote);
  db_write_head_raw(db, buf, len);
  unmap_memory(buf, len);

  return 0;
}

int gg_fetch(int argc, const char **argv, db_ptr db) {
  if (argc < 2) {
    die("Must supply source directory.");
  }
  const char *srcdir = argv[1];
  int slen = strlen(srcdir);
  char s[PATH_MAX];
  strcpy(s, srcdir);
  if (s[slen - 1] != '/') {
    s[slen] = '/';
    slen++;
    s[slen] ='\0';
  }
  strcpy(s + slen, "HEAD");
  if (!fileexists(s)) {
    strcpy(s + slen, ".gg/");
    slen += strlen(".gg/");
    strcpy(s + slen, "HEAD");
    if (!fileexists(s)) {
      die("Source not a gg directory");
    }
  }
  s[slen] = '\0';

  // Set source database.
  bchdir(s);
  if (!getcwd(s, PATH_MAX - 100)) die_perror("getcwd");
  slen = strlen(s);
  s[slen] = '/';
  slen++;
  s[slen] = '\0';
  db_ptr db_remote = db_new_gg(s);

  ID remote_id;
  db_read_head(remote_id, db_remote);

  // Exclude heads of branches in target repo.
  cbt_t exclude;
  cbt_init(exclude);
  void add_to_exclude(const char *ref) {
    ID local_id;
    db_read_branch(local_id, db, ref);
    cbt_put_u(exclude, (void *) 1, local_id, ID_LEN);
  }
  db_forall_branches(db, add_to_exclude);
  graph_t g;
  graph_init(g);
  graph_load_commit(g, db_remote, remote_id, exclude);
  cbt_clear(exclude);

  // Copy repo.
  void copy_tree(const ID id) {
    db_copy_object(db, db_remote, id);
    tree_t t;
    tree_init(t);
    tree_read(t, db_remote, id);
    void copy_tree_entry(void *p) {
      entry_ptr e = p;
      copy_tree(e->id);
    }
    darray_forall(t->tree_list, copy_tree_entry);
    void copy_blob_entry(void *p) {
      entry_ptr e = p;
      db_copy_object(db, db_remote, e->id);
    }
    darray_forall(t->blob_list, copy_blob_entry);
    tree_clear(t);
  }
  void copy_commit(commit_ptr c, const ID id) {
    db_copy_object(db, db_remote, id);
    copy_tree(c->tree_id);
    db_copy_object(db, db_remote, c->mode_id);
  }
  graph_forall(g, copy_commit);
  graph_clear(g);

  db_write_ref(db, "FETCH_HEAD", remote_id);

  db_delete(db_remote);
  return 0;
}

struct cmd_s {
  char *cmd;
  int (*fn)(int argc, const char **argv, db_ptr);
};

static struct cmd_s cmdtable[] = {
  { "init", gg_init_db },
  { "type", gg_type },
  { "cat", gg_cat },
  { "log", gg_log },

  { "add", gg_add },
  { "commit", gg_commit },
  { "checkout", gg_checkout },
  { "co", gg_checkout },
  { "save", gg_save },
  { "status", gg_status },

  { "export", gg_export },
  { "import", gg_import },
  { "git-import", gg_git_import },

  { "diff", gg_diff },
  { "fetch", gg_fetch },
  { "clone", gg_clone },
  { "lazy-clone", gg_lazy_clone },
  { NULL, NULL }
};

int main(int argc, const char **argv) {
  if (argc < 2) {
    usage();
  }
  // TODO: Overhaul required. Which commands need a working directory
  // and index? Which require an exisitng repository?
  char prefix[PATH_MAX];
  if (!getcwd(prefix, PATH_MAX - 100)) {
    die_perror("getcwd");
  }
  strcat(prefix, "/.gg/");
  char lazyfile[strlen(prefix) + 5];

  strcpy(lazyfile, prefix);
  strcat(lazyfile, "lazy");

  db_ptr db;
  if (fileexists(lazyfile)) {
    size_t len;
    char *buf = map_file_to_memory(&len, lazyfile);
    char *cp = strchr(buf, '\n');
    if (!cp) die("bad lazy file");
    char url[cp - buf + 1];
    strncpy(url, buf, cp - buf);
    unmap_memory(buf, len);
    url[cp - buf] = '\0';
    db = db_new_lazy(prefix, url);
  } else {
    db = db_new_gg(prefix);
  }

  struct cmd_s *c = cmdtable;
  int status;
  for (;;) {
    if (!c->cmd) {
      usage();
    }
    if (!strcmp(c->cmd, argv[1])) {
      status = c->fn(argc - 1, argv + 1, db);
      break;
    }
    c++;
  }

  db_delete(db);

  return status;
}
