// Import repository in git-fast-import format

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bstd.h"
#include "id.h"
#include "darray.h"
#include "db.h"
#include "tree.h"
#include "commit.h"
#include "cbt.h"
#include "fstree.h"
#include "list.h"

#include <stdint.h>
#include <inttypes.h>
#include "strbuf.h"
#include "quote.h"

static int initial_unless_explicit = 0;

// Copied liberally from git's fast-import.c.

struct object_entry_s {
	uint32_t type;
	// Since git-fast-export refers to blob by marks, it's best to
	// leave the SHA1 here and avoid a cbt lookup to get it.
	// It's a shame we use this for object_table as well, which has
	// no need for the SHA1.
	ID sha1;
};
typedef struct object_entry_s *object_entry_ptr;
typedef struct object_entry_s object_entry_t[1];

// git-fast-import keeps two versions of every file. I think they do this
// for packing purposes.
// I'm sticking to just one version for as long as I can get away with it.
struct branch_s {
  // We use fstree, not tree, because it sorts automatically.
  // Each fstree entry holds an ID, not an index_entry_t.
  fstree_t fstree;
  ID id;
  // Record executable files.
  size_t mode_blob_size;
  cbt_t xfiles;
};
// Need this function when recording mode information later.
void write_mode_blob_helper(db_ptr db, ID id,
      void (*forall_fn)(void (*)(void *)), size_t size);

typedef struct branch_s *branch_ptr;
typedef struct branch_s branch_t[1];

struct mark_set {
	union {
		object_entry_ptr marked[1024];
		struct mark_set *sets[1024];
	} data;
	unsigned int shift;
};

static const char *get_mode(const char *str, uint16_t *modep) {
	unsigned char c;
	uint16_t mode = 0;

	while ((c = *str++) != ' ') {
		if (c < '0' || c > '7')
			return NULL;
		mode = (mode << 3) + (c - '0');
	}
	*modep = mode;
	return str;
}

static cbt_t object_table;
static cbt_t branch_table;
static struct mark_set *marks;

struct recent_command {
  struct recent_command *prev;
  struct recent_command *next;
  char *buf;
};

static struct strbuf command_buf = STRBUF_INIT;
static int unread_command_buf;
static struct recent_command cmd_hist = {&cmd_hist, &cmd_hist, NULL};
static struct recent_command *cmd_tail = &cmd_hist;
static struct recent_command *rc_free;
static unsigned int cmd_save = 100;
static uintmax_t next_mark;

static branch_ptr lookup_branch(const char *name) {
  return cbt_at(branch_table, name);
}

static branch_ptr new_branch(const char *name)
{
  if (cbt_at(branch_table, name)) {
    die("Invalid attempt to create duplicate branch: %s", name);
  }

  /*
  switch (check_ref_format(name)) {
  case 0: break; // its valid
  case CHECK_REF_FORMAT_ONELEVEL:
	  break; // valid, but too few '/'
  default:
	  die("Branch name doesn't conform to GIT standards: %s", name);
  }
  */

  branch_ptr b = malloc(sizeof(branch_t));
  id_set0(b->id);
  fstree_init(b->fstree);
  cbt_init(b->xfiles);
  b->mode_blob_size = 3;

  cbt_put(branch_table, b, name);
  return b;
}

static void insert_mark(uintmax_t idnum, object_entry_ptr oe) {
	struct mark_set *s = marks;
	while ((idnum >> s->shift) >= 1024) {
		s = calloc(1, sizeof(struct mark_set));
		s->shift = marks->shift + 10;
		s->data.sets[0] = marks;
		marks = s;
	}
	while (s->shift) {
		uintmax_t i = idnum >> s->shift;
		idnum -= i << s->shift;
		if (!s->data.sets[i]) {
			s->data.sets[i] = calloc(1, sizeof(struct mark_set));
			s->data.sets[i]->shift = s->shift - 10;
		}
		s = s->data.sets[i];
	}
	//if (!s->data.marked[idnum])
		//marks_set_count++;
	s->data.marked[idnum] = oe;
}

static object_entry_ptr find_mark(uintmax_t idnum) {
	uintmax_t orig_idnum = idnum;
	struct mark_set *s = marks;
	object_entry_ptr oe = NULL;
	if ((idnum >> s->shift) < 1024) {
		while (s && s->shift) {
			uintmax_t i = idnum >> s->shift;
			idnum -= i << s->shift;
			s = s->data.sets[i];
		}
		if (s)
			oe = s->data.marked[idnum];
	}
	if (!oe)
		die("mark :%" PRIuMAX " not declared", orig_idnum);
	return oe;
}

static void recurse_clear_mark_set(struct mark_set *s) {
  if (s->shift) {
    int i = 0;
    while (i < 1024 && s->data.sets[i]) {
      recurse_clear_mark_set(s->data.sets[i]);
      i++;
    }
  }
  free(s);
}

static void clear_marks() {
  recurse_clear_mark_set(marks);
}

static object_entry_ptr new_object(ID sha1) {
  object_entry_ptr e = malloc(sizeof(object_entry_t));
  id_cpy(e->sha1, sha1);
  return e;
}

static object_entry_ptr find_object(ID sha1) {
  return cbt_at_u(object_table, sha1, 20);
}

static object_entry_ptr insert_object(ID sha1) {
  object_entry_ptr e = find_object(sha1);
  if (e) return e;
  e = new_object(sha1);
  cbt_put_u(object_table, e, sha1, 20);
  return e;
}

static void dump_branches(const db_ptr db) {
  void dump(void *p, const char *key) {
    branch_ptr b = p;
    db_write_ref(db, key, b->id);
  }
  cbt_forall(branch_table, dump);
}

static int read_next_command(void) {
  static int stdin_eof = 0;

  if (stdin_eof) {
    unread_command_buf = 0;
    return EOF;
  }

  do {
    if (unread_command_buf) {
      unread_command_buf = 0;
    } else {
      struct recent_command *rc;

      strbuf_detach(&command_buf, NULL);
      stdin_eof = strbuf_getline(&command_buf, stdin, '\n');
      if (stdin_eof)
        return EOF;

      rc = rc_free;
      if (rc)
        rc_free = rc->next;
      else {
        rc = cmd_hist.next;
        cmd_hist.next = rc->next;
        cmd_hist.next->prev = &cmd_hist;
        free(rc->buf);
      }

      rc->buf = command_buf.buf;
      rc->prev = cmd_tail;
      rc->next = cmd_hist.prev;
      rc->prev->next = rc;
      cmd_tail = rc;
    }
  } while (command_buf.buf[0] == '#');

  return 0;
}

static void skip_optional_lf(void) {
  int term_char = fgetc(stdin);
  if (term_char != '\n' && term_char != EOF) ungetc(term_char, stdin);
}


static void parse_mark(void) {
  if (!prefixcmp(command_buf.buf, "mark :")) {
    next_mark = strtoumax(command_buf.buf + 6, NULL, 10);
    read_next_command();
  } else next_mark = 0;
}

static void parse_data(struct strbuf *sb) {
  strbuf_reset(sb);

  if (prefixcmp(command_buf.buf, "data "))
    die("Expected 'data n' command, found: %s", command_buf.buf);

  if (!prefixcmp(command_buf.buf + 5, "<<")) {
	  char *term = bstrdup(command_buf.buf + 5 + 2);
	  size_t term_len = command_buf.len - 5 - 2;

	  // The detached buffer is freed later when it expires from
	  // the recent command circular buffer.
	  strbuf_detach(&command_buf, NULL);
	  for (;;) {
		  if (strbuf_getline(&command_buf, stdin, '\n') == EOF)
			  die("EOF in data (terminator '%s' not found)", term);
		  if (term_len == command_buf.len
			  && !strcmp(term, command_buf.buf))
			  break;
		  strbuf_addbuf(sb, &command_buf);
		  strbuf_addch(sb, '\n');
	  }
	  free(term);
  } else {
	  size_t n = 0, length;

	  length = strtoul(command_buf.buf + 5, NULL, 10);
	  while (n < length) {
		  size_t s = strbuf_fread(sb, length - n, stdin);
		  if (!s && feof(stdin))
			  die("EOF in data (%lu bytes remaining)",
				  (unsigned long)(length - n));
		  n += s;
	  }
  }

  skip_optional_lf();
}

static int validate_raw_date(const char *src, char *result, int maxlen)
{
  const char *orig_src = src;
  char *endp, sign;

  strtoul(src, &endp, 10);
  if (endp == src || *endp != ' ') return -1;

  src = endp + 1;
  if (*src != '-' && *src != '+') return -1;
  sign = *src;

  strtoul(src + 1, &endp, 10);
  if (endp == src || *endp || (endp - orig_src) >= maxlen) return -1;

  strcpy(result, orig_src);
  return 0;
}

static char *parse_ident(const char *buf) {
  const char *gt;
  size_t name_len;
  char *ident;

  gt = strrchr(buf, '>');
  if (!gt)
	  die("Missing > in ident string: %s", buf);
  gt++;
  if (*gt != ' ')
	  die("Missing space after > in ident string: %s", buf);
  gt++;
  name_len = gt - buf;
  ident = malloc(name_len + 24);
  strncpy(ident, buf, name_len);

  if (validate_raw_date(gt, ident + name_len, 24) < 0) {
    die("Invalid raw date \"%s\" in ident: %s", gt, buf);
  }

  return ident;
}

static void store_object(int type, struct strbuf *dat,
    db_ptr db, unsigned char *sha1out, uintmax_t mark) {
  object_entry_ptr e;
  ID sha1;

  id_header_buflen(sha1, typename(type), (unsigned char *) dat->buf, dat->len);
  if (sha1out) id_cpy(sha1out, sha1);

  e = insert_object(sha1);
  if (mark) insert_mark(mark, e);

  e->type = type;
  db_write_object(db, sha1, OBJ_BLOB, (unsigned char *) dat->buf, dat->len);
}

static struct strbuf blobbuf = STRBUF_INIT;
static void parse_new_blob(db_ptr db) {
  read_next_command();
  parse_mark();
  parse_data(&blobbuf);
  store_object(OBJ_BLOB, &blobbuf, db, NULL, next_mark);
}

static void fstree_put_slash_file(fstree_t fstree,
    ID id, const char *pathname) {
  char *cp;
  char *s, *dup = bstrdup(pathname);
  s = dup;
  fstree_ptr f = fstree;

  while ((cp = strchr(s, '/'))) {
    *cp = '\0';
    fstree_ptr fchild = cbt_at(f->child, s);
    if (!fchild) {
      fchild = malloc(sizeof(fstree_t));
      fstree_init(fchild);
      cbt_put(f->child, fchild, s);
    }
    f = fchild;
    s = cp + 1;
  }

  if (cbt_has(f->child, s)) {
    die("%s is dir and file!\n", s);
  }

  void *dup_if_null(void *p) {
    if (!p) return id_dup(id);
    id_cpy(p, id);
    return p;
  }
  cbt_put_with(f->entry, dup_if_null, s);
  free(dup);
}

static void fstree_remove_slash_file(fstree_t f, const char *pathname) {
  char *cp;
  char *dup = bstrdup(pathname);
  char *s = dup;

  while ((cp = strchr(s, '/'))) {
    *cp = '\0';
    fstree_ptr fchild = cbt_at(f->child, s);
    if (!fchild) {
      fprintf(stderr, "W: Deleting non-existent file: %s\n", pathname);
      return;
    }
    f = fchild;
    s = cp + 1;
  }

  if (cbt_at(f->entry, s)) {
    free(cbt_remove(f->entry, s));
  } else {
    fprintf(stderr, "W: Deleting non-existent file: %s\n", pathname);
  }
  free(dup);
}

static struct strbuf inline_data_buf = STRBUF_INIT;
static struct strbuf uqbuf = STRBUF_INIT;

static void file_change_m(branch_t b, db_ptr db) {
	const char *p = command_buf.buf + 2;
	const char *endp;
	object_entry_ptr oe = oe;
	ID sha1;
	uint16_t mode, inline_data = 0;

	p = get_mode(p, &mode);
	if (!p) die("Corrupt mode: %s", command_buf.buf);

	if (*p == ':') {
		char *x;
		oe = find_mark(strtoumax(p + 1, &x, 10));
		id_cpy(sha1, oe->sha1);
		p = x;
	} else if (!prefixcmp(p, "inline")) {
		inline_data = 1;
		p += 6;
	} else {
		id_sread(sha1, p);
		oe = find_object(sha1);
		p += 40;
	}
	if (*p++ != ' ')
		die("Missing space after SHA1: %s", command_buf.buf);

	strbuf_reset(&uqbuf);
	if (!unquote_c_style(&uqbuf, p, &endp)) {
		if (*endp)
			die("Garbage after path in: %s", command_buf.buf);
		p = uqbuf.buf;
	}
  if (mode == 0100755) {
    // Add to executables table.
    b->mode_blob_size += strlen(p) + 1;
    cbt_put(b->xfiles, (void *) 1, bstrdup(p));
  }

	if (inline_data) {
		if (p != uqbuf.buf) {
			strbuf_addstr(&uqbuf, p);
			p = uqbuf.buf;
		}
		read_next_command();
		parse_data(&inline_data_buf);
		store_object(OBJ_BLOB, &inline_data_buf, db, sha1, 0);
	} else if (oe) {
		if (oe->type != OBJ_BLOB)
			die("Not a blob (actually a %s): %s",
				typename(oe->type), command_buf.buf);
	} else {
		enum object_type type = db_stat(db, sha1, NULL);
		if (type < 0)
			die("Blob not found: %s", command_buf.buf);
		if (type != OBJ_BLOB)
			die("Not a blob (actually a %s): %s",
			    typename(type), command_buf.buf);
	}

	fstree_put_slash_file(b->fstree, sha1, p);
}

static void file_change_d(branch_t b) {
	const char *p = command_buf.buf + 2;
	const char *endp;

	strbuf_reset(&uqbuf);
	if (!unquote_c_style(&uqbuf, p, &endp)) {
		if (*endp)
			die("Garbage after path in: %s", command_buf.buf);
		p = uqbuf.buf;
	}
	fstree_remove_slash_file(b->fstree, p);

  if (cbt_has(b->xfiles, p)) {
    cbt_remove(b->xfiles, p);
    b->mode_blob_size -= strlen(p) + 1;
  }
}

static void fstree_del_recurse(fstree_t f) {
  fstree_clear_with(f, free);
}

static void reset_branch(branch_t b) {
  fstree_del_recurse(b->fstree);
  cbt_clear(b->xfiles);
  cbt_init(b->xfiles);
  b->mode_blob_size = 3;
}

static void file_change_deleteall(branch_ptr b) {
  reset_branch(b);
}

static void fstree_from_tree(fstree_ptr f, tree_ptr t, db_ptr db) {
  void read_tree(void *p) {
    entry_ptr e = p;
    tree_t tree;
    tree_init(tree);
    tree_read(tree, db, e->id);
    fstree_ptr fchild = malloc(sizeof(fstree_t));
    fstree_init(fchild);
    fstree_from_tree(fchild, tree, db);
    tree_clear(tree);
    cbt_put(f->child, fchild, e->s);
  }
  void read_blob(void *p) {
    entry_ptr e = p;
    // assert(!cbt_has(f->entry, e->s));
    cbt_put(f->entry, id_dup(e->id), e->s);
  }
  darray_forall(t->tree_list, read_tree);
  darray_forall(t->blob_list, read_blob);
}

// Load data from existing commit, which becomes the parent
// of the commit being parsed.
static void parse_from_existing(branch_ptr b, db_ptr db) {
  reset_branch(b);

  if (!id_is0(b->id)) {
    // TODO: Read only the beginning of the commit,
    // parse directly to fstree.
    commit_t commit;
    commit_init(commit);
    commit_read(commit, db, b->id);
    tree_t tree;
    tree_init(tree);
    tree_read(tree, db, commit->tree_id);
    commit_clear(commit);
    fstree_from_tree(b->fstree, tree, db);
    tree_clear(tree);
    void add(const char *filename) {
      cbt_put(b->xfiles, (void *) 1, filename);
    }
    b->mode_blob_size = forall_executables(db, commit->mode_id, add);
  }
}

static int parse_from(branch_ptr b, db_ptr db) {
  const char *from;
  branch_ptr s;

  // Normally if there is no "from" command and we're on an existing branch,
  // we start from the last known state of the branch as the parent.
  //
  // But this is problematic when there are distinct initial commits in
  // the same branch. With the current git-fast-export, you can tell which
  // these are because they're missing a "from" command.
  if (prefixcmp(command_buf.buf, "from ")) {
    if (initial_unless_explicit) {
      if (b) reset_branch(b);
      id_set0(b->id);
    }
    return 0;
  }

  if (b) reset_branch(b);

  from = strchr(command_buf.buf, ' ') + 1;
  s = lookup_branch(from);
  if (b == s) die("Can't create a branch from itself: %s", from);
  else if (s) {
    id_cpy(b->id, s->id);
    die("not supported yet!");
    // branch_copy(b, s);
  } else if (*from == ':') {
    uintmax_t idnum = strtoumax(from + 1, NULL, 10);
    object_entry_ptr oe = find_mark(idnum);
    if (oe->type != OBJ_COMMIT) die("Mark :%" PRIuMAX " not a commit", idnum);
    id_cpy(b->id, oe->sha1);
    parse_from_existing(b, db);
  //} else if (!get_sha1(from, b->sha1)) {
    //parse_from_existing(b);
  } else die("Invalid ref name or SHA1 expression: %s", from);

  read_next_command();
  return 1;
}

static void parse_merge(darray_t list) {
  const char *from;
  branch_ptr s;

  while (!prefixcmp(command_buf.buf, "merge ")) {
    from = strchr(command_buf.buf, ' ') + 1;
    s = lookup_branch(from);
    if (s) darray_append(list, id_dup(s->id));
    else if (*from == ':') {
      uintmax_t idnum = strtoumax(from + 1, NULL, 10);
      object_entry_ptr oe = find_mark(idnum);
      if (oe->type != OBJ_COMMIT) die("Mark :%" PRIuMAX " not a commit", idnum);
      darray_append(list, id_dup(oe->sha1));
    /* TODO
	} else if (!get_sha1(from, n->sha1)) {
			unsigned long size;
			char *buf = read_object_with_reference(n->sha1,
				commit_type, &size, n->sha1);
			if (!buf || size < 46)
				die("Not a valid commit: %s", from);
			free(buf);
    */
    } else die("Invalid ref name or SHA1 expression: %s", from);

    read_next_command();
  }
}

static struct strbuf msgbuf = STRBUF_INIT;

static void parse_new_commit(db_ptr db) {
  char *author = NULL;
  char *committer = NULL;

  // Obtain the branch name from the rest of our command
  char *sp = strchr(command_buf.buf, ' ') + 1;
  branch_ptr b = lookup_branch(sp);
  char *bname = strdup(sp);

  if (!b) b = new_branch(sp);
  read_next_command();
  parse_mark();
  if (!prefixcmp(command_buf.buf, "author ")) {
    author = parse_ident(command_buf.buf + 7);
    read_next_command();
  }
  if (!prefixcmp(command_buf.buf, "committer ")) {
    committer = parse_ident(command_buf.buf + 10);
    read_next_command();
  }
  if (!committer) die("Expected committer but didn't get one");

  parse_data(&msgbuf);
  read_next_command();
  parse_from(b, db);

  commit_t commit;
  commit_init(commit);
  if (!id_is0(b->id)) commit_add_parent(commit, b->id);

  parse_merge(commit->parent);

  // file_change*
  while (command_buf.len > 0) {
    if (!prefixcmp(command_buf.buf, "M ")) file_change_m(b, db);
    else if (!prefixcmp(command_buf.buf, "D ")) file_change_d(b);
    //else if (!prefixcmp(command_buf.buf, "R ")) file_change_cr(b, 1);
    //else if (!prefixcmp(command_buf.buf, "C ")) file_change_cr(b, 0);
    else if (!strcmp("deleteall", command_buf.buf)) {
      file_change_deleteall(b);
    } else {
      unread_command_buf = 1;
      break;
    }
    if (read_next_command() == EOF) break;
  }

/*
void fstree_print_keys(fstree_t root) {
  void recurse(void *p, const char *key) {
    printf("%s:\n", key);
    fstree_print_keys(p);
  }
  cbt_forall(root->child, recurse);
  void print(void *p, const char *key) {
    printf("%s\n", key);
  }
  cbt_forall(root->entry, print);
}
fstree_print_keys(b->fstree);
*/

  unsigned char *return_id(void *p) {
    entry_ptr e = p;
    return e->id;
  }
  write_tree_from_fstree(db, commit->tree_id, b->fstree, return_id);

  void forallx(void (*fn)(void *)) {
    void fnwrap(void *p, const char *key) {
      fn((void *) key);
    }
    cbt_forall(b->xfiles, fnwrap);
  }
  write_mode_blob_helper(db, commit->mode_id, forallx, b->mode_blob_size);

  commit->message = strdup((char *) msgbuf.buf);
  commit->author = author;
  commit->committer = committer;
  commit_write(db, b->id, commit);

  // TODO: Remove this line once branches work.
  db_write_head(db, b->id);

  free(bname);
  commit_clear(commit);

  // store_object(OBJ_COMMIT, &new_data, b->id, next_mark);
  object_entry_ptr e = insert_object(b->id);
  e->type = OBJ_COMMIT;
  if (next_mark) insert_mark(next_mark, e);
}

static void parse_new_tag(void) {
  static struct strbuf msg = STRBUF_INIT;
  char *sp;
  const char *from;
  char *tagger;
  branch_ptr s;
  // struct tag *t;
  uintmax_t from_mark = 0;
  unsigned char sha1[20];

  /* Obtain the new tag name from the rest of our command */
  sp = strchr(command_buf.buf, ' ') + 1;
	/*
	t = pool_alloc(sizeof(struct tag));
	t->next_tag = NULL;
	t->name = pool_strdup(sp);
	if (last_tag)
		last_tag->next_tag = t;
	else
		first_tag = t;
	last_tag = t;
	*/
  read_next_command();

  /* from ... */
  if (prefixcmp(command_buf.buf, "from "))
	  die("Expected from command, got %s", command_buf.buf);
  from = strchr(command_buf.buf, ' ') + 1;
  s = lookup_branch(from);
  if (s) {
    die("unsupported!\n");
    //id_cpy(sha1, s->sha1);
  } else if (*from == ':') {
    object_entry_ptr oe;
    from_mark = strtoumax(from + 1, NULL, 10);
    oe = find_mark(from_mark);
    if (oe->type != OBJ_COMMIT)
      die("Mark :%" PRIuMAX " not a commit", from_mark);
    id_cpy(sha1, oe->sha1);
    /*
  } else if (!get_sha1(from, sha1)) {
		unsigned long size;
		char *buf;

		buf = read_object_with_reference(sha1,
			commit_type, &size, sha1);
		if (!buf || size < 46)
			die("Not a valid commit: %s", from);
		free(buf);
		*/
  } else die("Invalid ref name or SHA1 expression: %s", from);
  read_next_command();

  // tagger
  if (prefixcmp(command_buf.buf, "tagger "))
	  die("Expected tagger command, got %s", command_buf.buf);
  tagger = parse_ident(command_buf.buf + 7);

  // tag payload/message
  read_next_command();
  parse_data(&msg);

  /*
  // build the tag object
  strbuf_reset(&new_data);
  strbuf_addf(&new_data,
	  "object %s\n"
	  "type %s\n"
	  "tag %s\n"
	  "tagger %s\n"
	  "\n",
	  sha1_to_hex(sha1), commit_type, t->name, tagger);
  strbuf_addbuf(&new_data, &msg);
  free(tagger);

  if (store_object(OBJ_TAG, &new_data, NULL, t->sha1, 0))
	  t->pack_id = MAX_PACK_ID;
  else
	  t->pack_id = pack_id;
	  */
}


static void parse_reset_branch(db_ptr db) {
  branch_ptr b;
  char *sp;

  // Obtain the branch name from the rest of our command
  sp = strchr(command_buf.buf, ' ') + 1;
  b = lookup_branch(sp);
  if (b) reset_branch(b);
  else b = new_branch(sp);
  read_next_command();
  parse_from(b, db);
  if (command_buf.len > 0) unread_command_buf = 1;
}

void import(db_ptr db, int explicit_parent_flag) {
  initial_unless_explicit = explicit_parent_flag;
  int i;
  marks = calloc(1, sizeof(struct mark_set));
  cbt_init(object_table);
  cbt_init(branch_table);
  strbuf_init(&command_buf, 0);

  struct recent_command *freeme = rc_free = malloc(cmd_save * sizeof(*rc_free));
  for (i = 0; i < (cmd_save - 1); i++) {
    rc_free[i].next = &rc_free[i + 1];
    rc_free[i].buf = NULL;
  }
  rc_free[cmd_save - 1].next = NULL;
  rc_free[cmd_save - 1].buf = NULL;
  while (read_next_command() != EOF) {
    if (!strcmp("blob", command_buf.buf)) parse_new_blob(db);
    else if (!prefixcmp(command_buf.buf, "commit ")) parse_new_commit(db);
    else if (!prefixcmp(command_buf.buf, "tag ")) parse_new_tag();
    else if (!prefixcmp(command_buf.buf, "reset ")) parse_reset_branch(db);
    else die("Unsupported command: '%s'", command_buf.buf);
  }
  dump_branches(db);
  void objfree(void *p, const char *key) {
    free(p);
  }
  cbt_clear_with(object_table, objfree);
  void branchfree(void *p, const char *key) {
    branch_ptr b = p;
    fstree_clear_with(b->fstree, free);
    cbt_clear(b->xfiles);
    free(b);
  }
  cbt_clear_with(branch_table, branchfree);

  // TODO: Figure out how to clear everything. Sometimes valgrind still
  // reports leaks.
  clear_marks();
  strbuf_release(&inline_data_buf);
  strbuf_release(&msgbuf);
  strbuf_release(&uqbuf);
  strbuf_release(&blobbuf);
  strbuf_release(&command_buf);

  struct recent_command *rc;
  for (rc = cmd_hist.next; rc != &cmd_hist; rc = rc->next) {
    free(rc->buf);
  }
  free(freeme);
}
