// Will inotify make all this stuff obsolete one day?

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <utime.h>
#include "darray.h"
#include "bstd.h"
#include "id.h"
#include "cbt.h"
#include "db.h"
#include "fstree.h"
#include "list.h"
#include "index.h"

#include <assert.h>

static int stat_cmp(index_entry_ptr ie, const struct stat *st) {
  int changed = 0;
  if (ie->ctime != (uint32_t) st->st_ctime) changed++;
  if (ie->mtime != (uint32_t) st->st_mtime) changed++;
  // if (ie->mode != (uint32_t) st->st_mode) changed++;
  if (ie->inode != (uint32_t) st->st_ino) changed++;
  if (ie->size != (uint32_t) st->st_size) changed++;
  return changed;
}

index_entry_ptr index_entry_dup(const index_entry_ptr ie) {
  index_entry_ptr res = malloc(sizeof(*res));
  *res = *ie;
  return res;
}

int index_entry_cmp(index_t index, index_entry_ptr ie, const char *pathname) {
  struct stat st[1];
  if (lstat(pathname, st)) die_perror("lstat %s", pathname);
  int changed = stat_cmp(ie, st);

  if (!changed && ie->dodgy) {
    ID id;
    size_t len;
    void *buf = map_file_to_memory(&len, pathname);
    id_header_buflen(id, "b", buf, len);
    unmap_memory(buf, len);
    if (id_cmp(ie->id, id)) changed++;
    else if (index->timestamp > ie->mtime) {
      // Clear dodgy flag to save future index operations
      // from recomputing the hash.
      ie->dodgy = 0;
    }
  }

  return changed;
}

static void fstree_add_path_recurse(time_t timestamp, db_ptr db,
    fstree_t f, const char *pathname);

static void fstree_add_dir(time_t timestamp, db_ptr db,
    fstree_t f, const char *dirname) {
  darray_t filelist;
  darray_init(filelist);

  void adddir(const char *filename) {
    darray_append(filelist, bstrdup(filename));
  }
  bopendir(dirname, adddir);
  bchdir(dirname);

  fstree_ptr fchild = cbt_at(f->child, dirname);
  if (!fchild) {
    fchild = malloc(sizeof(fstree_t));
    fstree_init(fchild);
    cbt_put(f->child, fchild, dirname);
  }
  void process(void *p) {
    char *filename = p;
    fstree_add_path_recurse(timestamp, db, fchild, filename);
  }
  darray_forall(filelist, process);
  darray_forall(filelist, free);
  darray_clear(filelist);
  bchdir("..");
}

static void slow_add_reg(time_t timestamp, db_ptr db, index_entry_ptr ie,
                  const char *regfile, const struct stat *st) {
  ie->ctime = st->st_ctime;
  ie->mtime = st->st_mtime;
  // Prevent race. Files just modified give dodgy hashes.
  // Even more so if they're from the future.
  if ((uint32_t) timestamp <= ie->mtime) ie->dodgy = 1;
  else ie->dodgy = 0;
  ie->mode = st->st_mode;
  ie->inode = st->st_ino;
  ie->size = st->st_size;

  size_t len;
  void *buf = map_file_to_memory(&len, regfile);
  id_header_buflen(ie->id, "b", buf, len);
  db_write_object(db, ie->id, OBJ_BLOB, buf, len);
  unmap_memory(buf, len);
}

static int index_entry_update(time_t timestamp, db_ptr db, index_entry_ptr ie,
                    const char *regfile, const struct stat *st) {
  if (stat_cmp(ie, st)) {
    slow_add_reg(timestamp, db, ie, regfile, st);
    return 1;
  }

  // Dodgy = can't trust hash.
  if (ie->dodgy) {
    ID id;
    size_t len;
    void *buf = map_file_to_memory(&len, regfile);
    id_header_buflen(ie->id, "b", buf, len);
    if (id_cmp(ie->id, id)) {
      db_write_object(db, id, OBJ_BLOB, buf, len);
      id_cpy(ie->id, id);
      unmap_memory(buf, len);
      return 1;
    } else {
      unmap_memory(buf, len);
      if (timestamp > ie->mtime) {
	// Clear dodgy flag to save future index operations
	// from recomputing the hash.
	ie->dodgy = 0;
      }
    }
  }

  // Otherwise trust hash and consider the file up-to-date.
  return 0;
}

static int fstree_add_reg(time_t timestamp, db_ptr db, fstree_ptr f,
                    const char *regfile, const struct stat *st) {
  index_entry_ptr ie = cbt_at(f->entry, regfile);
  if (!ie) {
    index_entry_ptr ie = malloc(sizeof(index_entry_t));
    slow_add_reg(timestamp, db, ie, regfile, st);
    cbt_put(f->entry, ie, regfile);
    return 1;
  }

  return index_entry_update(timestamp, db, ie, regfile, st);
}

static void fstree_add_path_recurse(time_t timestamp, db_ptr db,
    fstree_t f, const char *pathname) {
  assert(!strchr(pathname, '/'));
  struct stat st[1];
  if (lstat(pathname, st)) die_perror("lstat '%s'", pathname);
  if (S_ISREG(st->st_mode)) {
    fstree_add_reg(timestamp, db, f, pathname, st);
  } else if (S_ISDIR(st->st_mode)) {
    fstree_add_dir(timestamp, db, f, pathname);
  } else if (S_ISLNK(st->st_mode)) {
    fprintf(stderr, "%s: links not handled yet!\n", pathname);
  } else {
    fprintf(stderr,
	"skipping %s:"
	" can only add regular files, symbolic links or directories\n",
	pathname);
  }
}

// Add a path that may contain slashes.
static void fstree_add_slash_path(time_t timestamp, db_ptr db, fstree_t f,
    const char *pathname) {
  char *cp;
  // TODO: Remove recursion. Must chdir back to original directory at end.
  if ((cp = strchr(pathname, '/'))) {
    char s[cp - pathname + 1];
    strncpy(s, pathname, cp - pathname);
    s[cp - pathname] = '\0';
    bchdir(s);
    fstree_ptr fchild = cbt_at(f->child, s);
    if (!fchild) {
      fchild = malloc(sizeof(fstree_t));
      fstree_init(fchild);
      cbt_put(f->child, fchild, s);
    }
    fstree_add_slash_path(timestamp, db, fchild, cp + 1);
    bchdir("..");
  } else {
    fstree_add_path_recurse(timestamp, db, f, pathname);
  }
}

// Update a pathname if necessary. Not recursive.
// Returns whether the index entry was updated.
int fstree_update_path(index_ptr index, db_ptr db, fstree_ptr f,
    index_entry_ptr ie, const char *pathname) {
  struct stat st[1];
  if (lstat(pathname, st)) die_perror("lstat '%s'", pathname);
  int res = index_entry_update(index->timestamp, db, ie, pathname, st);
  cbt_put(f->entry, ie, pathname);
  return res;
}

void index_add_path(index_t index, db_ptr db, const char *fullname) {
  if (!index->has_read_timestamp) {
    // Return time according to the filesystem.
    // Touches a file and reads its mtime.
    // This defeats filesystem and system clock skew.
    index->timestamp = btouch_get_mtime(index->file);
    index->has_read_timestamp = 1;
  }
  fstree_add_slash_path(index->timestamp, db, index->root, fullname);
}

static void fstree_fprint(FILE *fp, fstree_t f) {
  void write_entry(void *p, const char *key) {
    index_entry_ptr ie = p;
    fprintf(fp, "%u ", strlen(key));
    id_fprint(fp, ie->id);
    fprintf(fp, " %x %x %x %x %x %s\n",
	    ie->ctime, ie->mtime, ie->mode, ie->inode, ie->size, key);
  }
  void write_child(void *p, const char *key) {
    fstree_ptr f = p;
    fprintf(fp, "%d %s\n", strlen(key), key);
    fstree_fprint(fp, f);
  }
  fprintf(fp, "%d\n", cbt_count(f->entry));
  cbt_forall(f->entry, write_entry);
  fprintf(fp, "%d\n", cbt_count(f->child));
  cbt_forall(f->child, write_child);
}

void index_write(const index_ptr index) {
  FILE *fp = bfopen(index->file, "wb");
  fstree_fprint(fp, index->root);
  bfclose(fp);
  struct utimbuf utim = { index->timestamp, index->timestamp };
  utime(index->file, &utim);
}

void fstree_read(fstree_ptr f, FILE *fp) {
  int i, entry_count, child_count;
  fscanf(fp, "%d\n", &entry_count);
  for (i = 0; i < entry_count; i++) {
    int n;
    fscanf(fp, "%d ", &n);
    index_entry_ptr ie = malloc(sizeof(index_entry_t));
    id_fread(ie->id, fp);
    if (5 != fscanf(fp, " %x %x %x %x %x ",
	            &ie->ctime, &ie->mtime, &ie->mode, &ie->inode, &ie->size)) {
      die("error reading index: fscanf");
    }
    char s[n + 1];
    fread(s, 1, n, fp);
    if ('\n' != fgetc(fp)) die("bad index\n");
    s[n] = '\0';
    cbt_put(f->entry, ie, s);
  }
  fscanf(fp, "%d\n", &child_count);
  for (i = 0; i < child_count; i++) {
    int n;
    fscanf(fp, "%d ", &n);
    char *dirname = malloc(n + 1);
    fread(dirname, 1, n, fp);
    dirname[n] = '\0';
    fstree_ptr fchild = malloc(sizeof(fstree_t));
    fstree_init(fchild);
    cbt_put(f->child, fchild, dirname);
    free(dirname);
    fstree_read(fchild, fp);
  }
}

void index_read(index_t index) {
  if (!fileexists(index->file)) return;
  FILE *fp = bfopen(index->file, "rb");
  fstree_read(index->root, fp);
  bfclose(fp);
}

void index_init(index_t index, const char *index_file) {
  fstree_init(index->root);
  index->has_read_timestamp = 0;
  index->file = strdup(index_file);
}

void index_clear(index_t index) {
  fstree_clear_with(index->root, free);
  free(index->file);
}
