#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bstd.h"
#include "id.h"
#include "darray.h"
#include "cbt.h"
#include "db.h"
#include "commit.h"
#include "list.h"
#include "graph.h"

struct node_s {
  int indegree;
  struct node_s **parent;
  commit_t commit;
  cbt_it it;
  int count;
};
typedef struct node_s *node_ptr;
typedef struct node_s node_t[1];

node_ptr new_node() {
  node_ptr node = malloc(sizeof(*node));
  node->indegree = 0;
  return node;
}

void graph_init(graph_t g) {
  cbt_init(g->node_table);
}

void graph_clear(graph_t g) {
  // TODO: Memory leaks: free node and node data.
  cbt_clear(g->node_table);
}

void graph_load_commit(graph_t g, db_ptr db, ID id, cbt_t exclude) {
  void recurse_load(ID id) {
    if (exclude && cbt_has_u(exclude, id, 20)) return;
    if (cbt_has_u(g->node_table, id, 20)) return;

    node_ptr node = new_node();
    commit_init(node->commit);
    commit_read(node->commit, db, id);
    node->it = cbt_put_u(g->node_table, node, id, 20);
    commit_forall_parent(node->commit, recurse_load);
    node->parent = malloc(
	commit_parent_count(node->commit) * sizeof(node_ptr));
    int i = 0;
    void makeedge(ID p_id) {
      node->parent[i] = (node_ptr) cbt_sure_at_u(g->node_table, p_id, 20);
      node->parent[i]->indegree++;
      i++;
    }
    commit_forall_parent(node->commit, makeedge);
  }
  recurse_load(id);
}

void topo_sort(list_ptr work, graph_ptr g) {
  // Topologically sort commits.
  int n = 0;
  void append(void *p, const char *key) {
    node_ptr node = p;
    if (!node->indegree) n++, list_append(work, node);
    else node->count = node->indegree;
  }
  cbt_forall(g->node_table, append);
  /* TODO: Optionally chronologically order commits.
  int datecmp(const void *p0, const void *p1) {
    return ((node_ptr) p0)->commit->date - ((node_ptr) p1)->commit->date;
  }
  darray_qsort(sortme, datecmp);
  */
  list_it lit = list_it_first(work);
  while(!list_it_is_off(lit)) {
    node_ptr node = list_it_data(lit);
    int i;
    for (i = 0; i < commit_parent_count(node->commit); i++) {
      node_ptr parent = node->parent[i];
      parent->count--;
      if (!parent->count) n++, list_append(work, parent);
    }
    lit = list_it_next(lit);
  }
}

static void show_log(db_ptr db, ID start_id) {
  // Load commits into graph.
  graph_t g;
  graph_init(g);
  graph_load_commit(g, db, start_id, NULL);

  list_t work;
  list_init(work);
  topo_sort(work, g);

  int first = 1;
  void print(void *p) {
    node_ptr node = p;
    commit_ptr commit = node->commit;
    if (first) {
      first = 0;
    } else {
      printf("\n");
    }
    printf("commit ");
    id_fprint(stdout, (unsigned char *) cbt_it_key(node->it));
    printf("\nAuthor: %s\n",
      commit->author ? commit->author : commit->committer);
    void print_parent(ID p_id) {
      printf("parent: ");
      id_fprint(stdout, p_id);
      printf("\n");
    }
    commit_forall_parent(commit, print_parent);
    printf("tree: ");
    id_fprint(stdout, commit->tree_id);
    printf("\n\n");
    char *c = commit->message;
    char *end = c + commit->messagelen;
    int mustindent = 1;
    while (c != end) {
      if (mustindent) {
	fputs("    ", stdout);
	mustindent = 0;
      }
      if ('\n' == *c) {
	mustindent = 1;
      }
      putchar(*c);
      c++;
    }
  }
  list_forall(work, print);
  // TODO: Fix memory leaks.
  list_clear(work);
  graph_clear(g);
}

int gg_log(int argc, const char **argv, db_ptr db) {
  ID id;
  if (argc < 2) {
    db_read_head(id, db);
  } else {
    parse_rev(id, db, argv[1]);
  }
  if (id_is0(id)) {
    printf("No commits yet.\n");
    return 0;
  }
  show_log(db, id);

  return 0;
}

void graph_forall(graph_t g, void (*fn)(commit_ptr c, const ID id)) {
  void wrap(void *data, const char *key) {
    node_ptr p = data;
    fn(p->commit, (const unsigned char *) key);
  }
  cbt_forall(g->node_table, wrap);
}
