// Trees: lists of trees or blobs

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bstd.h"
#include "id.h"
#include "darray.h"
#include "db.h"
#include "tree.h"

void entry_init(entry_t e, const ID id, const char *s) {
  e->s = strdup(s);
  id_cpy(e->id, id);
}

void entry_clear(entry_t e) {
  free(e->s);
}

void tree_init(tree_t tree) {
  darray_init(tree->tree_list);
  darray_init(tree->blob_list);
}

void tree_clear(tree_t tree) {
  void clear(void *p) { entry_clear(p); free(p); }
  darray_forall(tree->blob_list, clear);
  darray_forall(tree->tree_list, clear);
  darray_clear(tree->tree_list);
  darray_clear(tree->blob_list);
}

void tree_append_blob(tree_t tree, const ID id, const char *s) {
  // Assumes neither list already contains the string 's'
  entry_ptr e = malloc(sizeof(entry_t));
  entry_init(e, id, s);
  darray_append(tree->blob_list, e);
}

void tree_append_tree(tree_t tree, const ID id, const char *s) {
  // Assumes neither list already contains the string 's'
  entry_ptr e = malloc(sizeof(entry_t));
  entry_init(e, id, s);
  darray_append(tree->tree_list, e);
}

static void tree_read_body(tree_t tree, char *buf, size_t len) {
  char *cp = buf;
  void (*adder)(tree_t, const ID, const char *) = tree_append_tree;
  while (cp < buf + len) {
    if ('\n' == *cp) {
      cp++;
      if (adder != tree_append_tree) {
	die("invalid tree: extra newline");
      } else {
	adder = tree_append_blob;
      }
    }
    if ('\0' == *cp) {
      // Special case: no files in directory.
      // We insist on the newline separating the files and directories however.
      if (adder != tree_append_blob) {
	die("invalid tree: missing newline");
      }
      if (buf + len - cp) {
	die("invalid tree: extra characters on end\n");
      }
      break;
    }
    ID id;
    id_sread(id, cp);
    cp += 40;
    if (' ' != *cp) die("invalid tree: missing space after hash");
    cp++;
    char *np = strchr(cp, '\n');
    *np = '\0';
    adder(tree, id, cp);
    cp = np + 1;
  }
}

void tree_read(tree_t tree, db_ptr db, const ID id) {
  size_t bodylen;
  void *bodybuf = db_mmap_body_check_type(&bodylen, OBJ_TREE, db, id);
  tree_read_body(tree, bodybuf, bodylen);
  db_munmap(bodybuf, bodylen);
}

void compute_id_and_write_tree(db_ptr db, ID id, tree_t tree) {
  int tree_size = 2;  // For newline and terminator.
  void account_for_entry(void *p) {
    entry_ptr e = p;
    tree_size += strlen(e->s) + 42;  // String, hash, space and newline.
  }
  darray_forall(tree->tree_list, account_for_entry);
  darray_forall(tree->blob_list, account_for_entry);

  char s[tree_size];
  *s = '\0';
  void append(void *p) {
    entry_ptr e = p;
    id_sprint(s + strlen(s), e->id);
    strcat(s, " ");
    strcat(s, e->s);
    strcat(s, "\n");
  }
  darray_forall(tree->tree_list, append);
  strcat(s, "\n");
  darray_forall(tree->blob_list, append);
  id_header_string(id, "t", s);
  db_write_object(db, id, OBJ_TREE, (unsigned char *) s, strlen(s));
}
